from pydantic import BaseModel, Field
import nodetool.metadata.types
from nodetool.metadata.types import *
from nodetool.dsl.graph import GraphNode

from nodetool.nodes.huggingface.image.classify import ModelId

class Classifier(GraphNode):
    model: ModelId | GraphNode | tuple[GraphNode, str] = Field(default=ModelId('google/vit-base-patch16-224'), description='The model ID to use for the classification')
    image: ImageRef | GraphNode | tuple[GraphNode, str] = Field(default=ImageRef(type='image', uri='', asset_id=None), description='The input image to classify')
    @classmethod
    def get_node_type(cls): return "huggingface.image.classify.Classifier"





from nodetool.api.server import create_app

# in local or dev mode, import all nodes
import nodetool.nodes.anthropic
import nodetool.nodes.comfy
import nodetool.nodes.huggingface
import nodetool.nodes.nodetool
import nodetool.nodes.openai
import nodetool.nodes.replicate

app = create_app()



# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
# Copyright (c) @comfyanonymous
# Project Repository: https://github.com/comfyanonymous/ComfyUI

import comfy.folder_paths as folder_paths
import comfy.sd
import comfy.model_sampling
import torch

class LCM(comfy.model_sampling.EPS):
    def calculate_denoised(self, sigma, model_output, model_input):
        timestep = self.timestep(sigma).view(sigma.shape[:1] + (1,) * (model_output.ndim - 1))
        sigma = sigma.view(sigma.shape[:1] + (1,) * (model_output.ndim - 1))
        x0 = model_input - model_output * sigma

        sigma_data = 0.5
        scaled_timestep = timestep * 10.0 #timestep_scaling

        c_skip = sigma_data**2 / (scaled_timestep**2 + sigma_data**2)
        c_out = scaled_timestep / (scaled_timestep**2 + sigma_data**2) ** 0.5

        return c_out * x0 + c_skip * model_input

class ModelSamplingDiscreteDistilled(comfy.model_sampling.ModelSamplingDiscrete):
    original_timesteps = 50

    def __init__(self, model_config=None):
        super().__init__(model_config)

        self.skip_steps = self.num_timesteps // self.original_timesteps

        sigmas_valid = torch.zeros((self.original_timesteps), dtype=torch.float32)
        for x in range(self.original_timesteps):
            sigmas_valid[self.original_timesteps - 1 - x] = self.sigmas[self.num_timesteps - 1 - x * self.skip_steps]

        self.set_sigmas(sigmas_valid)

    def timestep(self, sigma):
        log_sigma = sigma.log()
        dists = log_sigma.to(self.log_sigmas.device) - self.log_sigmas[:, None]
        return (dists.abs().argmin(dim=0).view(sigma.shape) * self.skip_steps + (self.skip_steps - 1)).to(sigma.device)

    def sigma(self, timestep):
        t = torch.clamp(((timestep.float().to(self.log_sigmas.device) - (self.skip_steps - 1)) / self.skip_steps).float(), min=0, max=(len(self.sigmas) - 1))
        low_idx = t.floor().long()
        high_idx = t.ceil().long()
        w = t.frac()
        log_sigma = (1 - w) * self.log_sigmas[low_idx] + w * self.log_sigmas[high_idx]
        return log_sigma.exp().to(timestep.device)


def rescale_zero_terminal_snr_sigmas(sigmas):
    alphas_cumprod = 1 / ((sigmas * sigmas) + 1)
    alphas_bar_sqrt = alphas_cumprod.sqrt()

    # Store old values.
    alphas_bar_sqrt_0 = alphas_bar_sqrt[0].clone()
    alphas_bar_sqrt_T = alphas_bar_sqrt[-1].clone()

    # Shift so the last timestep is zero.
    alphas_bar_sqrt -= (alphas_bar_sqrt_T)

    # Scale so the first timestep is back to the old value.
    alphas_bar_sqrt *= alphas_bar_sqrt_0 / (alphas_bar_sqrt_0 - alphas_bar_sqrt_T)

    # Convert alphas_bar_sqrt to betas
    alphas_bar = alphas_bar_sqrt**2  # Revert sqrt
    alphas_bar[-1] = 4.8973451890853435e-08
    return ((1 - alphas_bar) / alphas_bar) ** 0.5

class ModelSamplingDiscrete:
    @classmethod
    def INPUT_TYPES(s):
        return {"required": { "model": ("MODEL",),
                              "sampling": (["eps", "v_prediction", "lcm"],),
                              "zsnr": ("BOOLEAN", {"default": False}),
                              }}

    RETURN_TYPES = ("MODEL",)
    FUNCTION = "patch"

    CATEGORY = "advanced/model"

    def patch(self, model, sampling, zsnr):
        m = model.clone()

        sampling_base = comfy.model_sampling.ModelSamplingDiscrete
        if sampling == "eps":
            sampling_type = comfy.model_sampling.EPS
        elif sampling == "v_prediction":
            sampling_type = comfy.model_sampling.V_PREDICTION
        elif sampling == "lcm":
            sampling_type = LCM
            sampling_base = ModelSamplingDiscreteDistilled

        class ModelSamplingAdvanced(sampling_base, sampling_type):
            pass

        model_sampling = ModelSamplingAdvanced(model.model.model_config)
        if zsnr:
            model_sampling.set_sigmas(rescale_zero_terminal_snr_sigmas(model_sampling.sigmas))

        m.add_object_patch("model_sampling", model_sampling)
        return (m, )

class ModelSamplingContinuousEDM:
    @classmethod
    def INPUT_TYPES(s):
        return {"required": { "model": ("MODEL",),
                              "sampling": (["v_prediction", "eps"],),
                              "sigma_max": ("FLOAT", {"default": 120.0, "min": 0.0, "max": 1000.0, "step":0.001, "round": False}),
                              "sigma_min": ("FLOAT", {"default": 0.002, "min": 0.0, "max": 1000.0, "step":0.001, "round": False}),
                              }}

    RETURN_TYPES = ("MODEL",)
    FUNCTION = "patch"

    CATEGORY = "advanced/model"

    def patch(self, model, sampling, sigma_max, sigma_min):
        m = model.clone()

        if sampling == "eps":
            sampling_type = comfy.model_sampling.EPS
        elif sampling == "v_prediction":
            sampling_type = comfy.model_sampling.V_PREDICTION

        class ModelSamplingAdvanced(comfy.model_sampling.ModelSamplingContinuousEDM, sampling_type):
            pass

        model_sampling = ModelSamplingAdvanced(model.model.model_config)
        model_sampling.set_sigma_range(sigma_min, sigma_max)
        m.add_object_patch("model_sampling", model_sampling)
        return (m, )

class RescaleCFG:
    @classmethod
    def INPUT_TYPES(s):
        return {"required": { "model": ("MODEL",),
                              "multiplier": ("FLOAT", {"default": 0.7, "min": 0.0, "max": 1.0, "step": 0.01}),
                              }}
    RETURN_TYPES = ("MODEL",)
    FUNCTION = "patch"

    CATEGORY = "advanced/model"

    def patch(self, model, multiplier):
        def rescale_cfg(args):
            cond = args["cond"]
            uncond = args["uncond"]
            cond_scale = args["cond_scale"]
            sigma = args["sigma"]
            sigma = sigma.view(sigma.shape[:1] + (1,) * (cond.ndim - 1))
            x_orig = args["input"]

            #rescale cfg has to be done on v-pred model output
            x = x_orig / (sigma * sigma + 1.0)
            cond = ((x - (x_orig - cond)) * (sigma ** 2 + 1.0) ** 0.5) / (sigma)
            uncond = ((x - (x_orig - uncond)) * (sigma ** 2 + 1.0) ** 0.5) / (sigma)

            #rescalecfg
            x_cfg = uncond + cond_scale * (cond - uncond)
            ro_pos = torch.std(cond, dim=(1,2,3), keepdim=True)
            ro_cfg = torch.std(x_cfg, dim=(1,2,3), keepdim=True)

            x_rescaled = x_cfg * (ro_pos / ro_cfg)
            x_final = multiplier * x_rescaled + (1.0 - multiplier) * x_cfg

            return x_orig - (x - x_final * sigma / (sigma * sigma + 1.0) ** 0.5)

        m = model.clone()
        m.set_model_sampler_cfg_function(rescale_cfg)
        return (m, )

NODE_CLASS_MAPPINGS = {
    "ModelSamplingDiscrete": ModelSamplingDiscrete,
    "ModelSamplingContinuousEDM": ModelSamplingContinuousEDM,
    "RescaleCFG": RescaleCFG,
}



# MangaLineExtraction_PyTorch
# https://github.com/ljsabc/MangaLineExtraction_PyTorch

#NOTE: This preprocessor is designed to work with lineart_anime ControlNet so the result will be white lines on black canvas

import torch
import numpy as np
import os
import cv2
from einops import rearrange
from .model_torch import res_skip
from PIL import Image
import warnings

from controlnet_aux.util import HWC3, resize_image_with_pad, common_input_validate, custom_hf_download, HF_MODEL_NAME

class LineartMangaDetector:
    def __init__(self, model):
        self.model = model

    @classmethod
    def from_pretrained(cls, pretrained_model_or_path=HF_MODEL_NAME, filename="erika.pth"):
        model_path = custom_hf_download(pretrained_model_or_path, filename)

        net = res_skip()
        ckpt = torch.load(model_path)
        for key in list(ckpt.keys()):
            if 'module.' in key:
                ckpt[key.replace('module.', '')] = ckpt[key]
                del ckpt[key]
        net.load_state_dict(ckpt)
        net.eval()
        return cls(net)

    def to(self, device):
        self.model.to(device)
        return self

    def __call__(self, input_image, detect_resolution=512, output_type="pil", upscale_method="INTER_CUBIC", **kwargs):
        input_image, output_type = common_input_validate(input_image, output_type, **kwargs)
        detected_map, remove_pad = resize_image_with_pad(input_image, 256 * int(np.ceil(float(detect_resolution) / 256.0)), upscale_method)
        device = next(iter(self.model.parameters())).device

        img = cv2.cvtColor(detected_map, cv2.COLOR_RGB2GRAY)
        with torch.no_grad():
            image_feed = torch.from_numpy(img).float().to(device)
            image_feed = rearrange(image_feed, 'h w -> 1 1 h w')

            line = self.model(image_feed)
            line = line.cpu().numpy()[0,0,:,:]
            line[line > 255] = 255
            line[line < 0] = 0

            line = line.astype(np.uint8)
        
        detected_map = HWC3(line)
        detected_map = remove_pad(255 - detected_map)
        
        if output_type == "pil":
            detected_map = Image.fromarray(detected_map)
            
        return detected_map



from enum import Enum
from typing import Literal
from pydantic import Field
from nodetool.common.huggingface_node import HuggingfaceNode
from nodetool.metadata.types import BaseType
from nodetool.workflows.processing_context import ProcessingContext


class ModelId(str, Enum):
    CARDIFFNLP_TWITTER_ROBERTA_BASE_SENTIMENT_LATEST = (
        "cardiffnlp/twitter-roberta-base-sentiment-latest"
    )
    J_HARTMANN_EMOTION_ENGLISH_DISTILROBERTA_BASE = (
        "j-hartmann/emotion-english-distilroberta-base"
    )
    SAMLOWE_ROBERTA_BASE_GO_EMOTIONS = "SamLowe/roberta-base-go_emotions"
    PROSUSAI_FINBERT = "ProsusAI/finbert"
    DISTILBERT_BASE_UNCASED_FINETUNED_SST_2_ENGLISH = (
        "distilbert/distilbert-base-uncased-finetuned-sst-2-english"
    )


class Classifier(HuggingfaceNode):
    """
    Text Classification is the task of assigning a label or class to a given text. Some use cases are sentiment analysis, natural language inference, and assessing grammatical correctness.
    classifier, text, huggingface

    ### Use Cases
    * You can track the sentiments of your customers from the product reviews using sentiment analysis models. This can help understand churn and retention by grouping reviews by sentiment, to later analyze the text and make strategic decisions based on this knowledge.
    """

    model: ModelId = Field(
        default=ModelId.CARDIFFNLP_TWITTER_ROBERTA_BASE_SENTIMENT_LATEST,
        title="Model ID on Huggingface",
        description="The model ID to use for the classification",
    )
    inputs: str = Field(
        default="",
        title="Inputs",
        description="The input text to classify",
    )

    async def process(self, context: ProcessingContext) -> dict[str, float]:
        result = await self.run_huggingface(
            model_id=self.model.value, context=context, params={"inputs": self.inputs}
        )
        return result[0]



from ..utils import common_annotator_call, create_node_input_types
import comfy.model_management as model_management
import torch
from einops import rearrange

class AnimeFace_SemSegPreprocessor:
    @classmethod
    def INPUT_TYPES(s):
        return {
            "required": {
                "image": ("IMAGE",)
            },
            "optional": {
                #This preprocessor is only trained on 512x resolution
                #https://github.com/siyeong0/Anime-Face-Segmentation/blob/main/predict.py#L25
                "remove_background_using_abg": ("BOOLEAN", {"default": True}),
                "resolution": ("INT", {"default": 512, "min": 512, "max": 512, "step": 64})
            }
        }

    RETURN_TYPES = ("IMAGE", "MASK")
    RETURN_NAMES = ("IMAGE", "ABG_CHARACTER_MASK (MASK)")
    FUNCTION = "execute"

    CATEGORY = "ControlNet Preprocessors/Semantic Segmentation"

    def execute(self, image, remove_background_using_abg=True, resolution=512, **kwargs):
        from controlnet_aux.anime_face_segment import AnimeFaceSegmentor

        model = AnimeFaceSegmentor.from_pretrained().to(model_management.get_torch_device())
        if remove_background_using_abg:
            out_image_with_mask = common_annotator_call(model, image, resolution=resolution, remove_background=True)
            out_image = out_image_with_mask[..., :3]
            mask = out_image_with_mask[..., 3:]
            mask = rearrange(mask, "n h w c -> n c h w")
        else:
            out_image = common_annotator_call(model, image, resolution=resolution, remove_background=False)
            N, H, W, C = out_image.shape
            mask = torch.ones(N, C, H, W)
        del model
        return (out_image, mask)

NODE_CLASS_MAPPINGS = {
    "AnimeFace_SemSegPreprocessor": AnimeFace_SemSegPreprocessor
}
NODE_DISPLAY_NAME_MAPPINGS = {
    "AnimeFace_SemSegPreprocessor": "Anime Face Segmentor"
}


from enum import Enum
from pydantic import Field
from nodetool.metadata.types import Conditioning, Latent, UNet
from nodetool.common.comfy_node import EnableDisable
from nodetool.common.comfy_node import ComfyNode


class SamplerEnum(str, Enum):
    DDIM = "ddim"
    UNI_PC = "uni_pc"
    UNI_PC_BH2 = "uni_pc_bh2"
    EULER = "euler"
    EULER_ANCESTRAL = "euler_ancestral"
    HEUN = "heun"
    HEUNPP2 = "heunpp2"
    DPM_2 = "dpm_2"
    DPM_2_ANCESTRAL = "dpm_2_ancestral"
    LMS = "lms"
    DPM_FAST = "dpm_fast"
    DPM_ADAPTIVE = "dpm_adaptive"
    DPMPP_2S_ANCESTRAL = "dpmpp_2s_ancestral"
    DPMPP_SDE = "dpmpp_sde"
    DPMPP_SDE_GPU = "dpmpp_sde_gpu"
    DPMPP_2M = "dpmpp_2m"
    DPMPP_2M_SDE = "dpmpp_2m_sde"
    DPMPP_2M_SDE_GPU = "dpmpp_2m_sde_gpu"
    DPMPP_3M_SDE = "dpmpp_3m_sde"
    DPMPP_3M_SDE_GPU = "dpmpp_3m_sde_gpu"
    DDPM = "ddpm"
    LCM = "lcm"


class SchedulerEnum(str, Enum):
    NORMAL = "normal"
    KARRAS = "karras"
    EXPONENTIAL = "exponential"
    SGM_UNIFORM = "sgm_uniform"
    SIMPLE = "simple"
    DDIM_UNIFORM = "ddim_uniform"


class KSampler(ComfyNode):
    """
    The KSampler uses the provided model and positive and negative conditioning to generate a new version of the given latent. First the latent is noised up according to the given seed and denoise strength, erasing some of the latent image. then this noise is removed using the given Model and the positive and negative conditioning as guidance, "dreaming" up new details in places where the image was erased by noise.
    """

    model: UNet = Field(default=UNet(), description="The model to use.")
    seed: int = Field(default=0, description="The seed to use.")
    steps: int = Field(default=20, description="The number of steps to use.")
    cfg: float = Field(default=8.0, description="The cfg to use.")
    sampler_name: SamplerEnum = Field(
        default=SamplerEnum.DDIM, description="The sampler to use."
    )
    scheduler: SchedulerEnum = Field(
        default=SchedulerEnum.NORMAL, description="The scheduler to use."
    )
    positive: Conditioning = Field(
        default=Conditioning(), description="The positive conditioning to use."
    )
    negative: Conditioning = Field(
        default=Conditioning(), description="The negative conditioning to use."
    )
    latent_image: Latent = Field(
        default=Latent(), description="The latent image to use."
    )
    denoise: float = Field(default=1.0, description="The denoise to use.")

    @classmethod
    def return_type(cls):
        return {"latent": Latent}


class KSamplerAdvanced(ComfyNode):
    """
    The KSampler Advanced node is the more advanced version of the KSampler node. While the KSampler node always adds noise to the latent followed by completely denoising the noised up latent, the KSampler Advanced node provides extra settings to control this behavior. The KSampler Advanced node can be told not to add noise into the latent with the add_noise setting. It can also be made to return partially denoised images via the return_with_leftover_noise setting. Unlike the KSampler node, this node does not have a denoise setting but this process is instead controlled by the start_at_step and end_at_step settings. This makes it possible to e.g. hand over a partially denoised latent to a separate KSampler Advanced node to finish the process.
    """

    model: UNet = Field(default=UNet(), description="The model to use.")
    add_noise: EnableDisable = Field(
        default=EnableDisable.ENABLE, description="Enable or disable noise addition."
    )
    noise_seed: int = Field(default=0, description="The seed for noise generation.")
    steps: int = Field(
        default=20, description="The number of steps to use during sampling."
    )
    cfg: float = Field(
        default=8.0, description="The configuration value for the sampler."
    )
    sampler_name: SamplerEnum = Field(
        default=SamplerEnum.DDIM, description="The name of the sampler to use."
    )
    scheduler: SchedulerEnum = Field(
        default=SchedulerEnum.NORMAL, description="The scheduler to use."
    )
    positive: Conditioning = Field(
        default=Conditioning(), description="The positive conditioning influence."
    )
    negative: Conditioning = Field(
        default=Conditioning(), description="The negative conditioning influence."
    )
    latent_image: Latent = Field(
        default=Latent(), description="The starting latent image."
    )
    start_at_step: int = Field(
        default=0, description="The step at which to start the sampling process."
    )
    end_at_step: int = Field(
        default=10000, description="The step at which to end the sampling process."
    )
    return_with_leftover_noise: EnableDisable = Field(
        default=EnableDisable.DISABLE,
        description="Whether to return with leftover noise or not.",
    )

    @classmethod
    def return_type(cls):
        return {"latent": Latent}



from ..utils import common_annotator_call, create_node_input_types


class Tile_Preprocessor:
    @classmethod
    def INPUT_TYPES(s):
        return create_node_input_types(
            pyrUp_iters = ("INT", {"default": 3, "min": 1, "max": 10, "step": 1})
        )
        

    RETURN_TYPES = ("IMAGE",)
    FUNCTION = "execute"

    CATEGORY = "ControlNet Preprocessors/others"

    def execute(self, image, pyrUp_iters, resolution=512, **kwargs):
        from controlnet_aux.tile import TileDetector

        return (common_annotator_call(TileDetector(), image, pyrUp_iters=pyrUp_iters, resolution=resolution),)


NODE_CLASS_MAPPINGS = {
    "TilePreprocessor": Tile_Preprocessor,
}

NODE_DISPLAY_NAME_MAPPINGS = {
    "TilePreprocessor": "Tile"
}



from pydantic import BaseModel, Field
import nodetool.metadata.types
from nodetool.metadata.types import *
from nodetool.dsl.graph import GraphNode


class ConcatAudio(GraphNode):
    a: AudioRef | GraphNode | tuple[GraphNode, str] = Field(default=AudioRef(type='audio', uri='', asset_id=None, temp_id=None), description='The first audio file.')
    b: AudioRef | GraphNode | tuple[GraphNode, str] = Field(default=AudioRef(type='audio', uri='', asset_id=None, temp_id=None), description='The second audio file.')
    @classmethod
    def get_node_type(cls): return "nodetool.audio.transform.ConcatAudio"



class NormalizeAudio(GraphNode):
    audio: AudioRef | GraphNode | tuple[GraphNode, str] = Field(default=AudioRef(type='audio', uri='', asset_id=None, temp_id=None), description='The audio file to normalize.')
    @classmethod
    def get_node_type(cls): return "nodetool.audio.transform.NormalizeAudio"



class OverlayAudio(GraphNode):
    a: AudioRef | GraphNode | tuple[GraphNode, str] = Field(default=AudioRef(type='audio', uri='', asset_id=None, temp_id=None), description='The first audio file.')
    b: AudioRef | GraphNode | tuple[GraphNode, str] = Field(default=AudioRef(type='audio', uri='', asset_id=None, temp_id=None), description='The second audio file.')
    @classmethod
    def get_node_type(cls): return "nodetool.audio.transform.OverlayAudio"



class RemoveSilence(GraphNode):
    audio: AudioRef | GraphNode | tuple[GraphNode, str] = Field(default=AudioRef(type='audio', uri='', asset_id=None, temp_id=None), description='The audio file to remove silence from.')
    @classmethod
    def get_node_type(cls): return "nodetool.audio.transform.RemoveSilence"



class SliceAudio(GraphNode):
    audio: AudioRef | GraphNode | tuple[GraphNode, str] = Field(default=AudioRef(type='audio', uri='', asset_id=None, temp_id=None), description='The audio file.')
    start: float | GraphNode | tuple[GraphNode, str] = Field(default=0.0, description='The start time in seconds.')
    end: float | GraphNode | tuple[GraphNode, str] = Field(default=1.0, description='The end time in seconds.')
    @classmethod
    def get_node_type(cls): return "nodetool.audio.transform.SliceAudio"



class Tone(GraphNode):
    frequency: float | GraphNode | tuple[GraphNode, str] = Field(default=440.0, description='Frequency of the tone in Hertz.')
    sampling_rate: int | GraphNode | tuple[GraphNode, str] = Field(default=44100, description='Sampling rate.')
    duration: float | GraphNode | tuple[GraphNode, str] = Field(default=1.0, description='Duration of the tone in seconds.')
    phi: float | GraphNode | tuple[GraphNode, str] = Field(default=0.0, description='Initial phase of the waveform in radians.')
    @classmethod
    def get_node_type(cls): return "nodetool.audio.transform.Tone"





from pydantic import BaseModel, Field
import nodetool.metadata.types
from nodetool.metadata.types import *
from nodetool.dsl.graph import GraphNode


class CodeFormer(GraphNode):
    image: ImageRef | GraphNode | tuple[GraphNode, str] = Field(default=ImageRef(type='image', uri='', asset_id=None), description='Input image')
    upscale: int | GraphNode | tuple[GraphNode, str] = Field(default=2, description='The final upsampling scale of the image')
    face_upsample: bool | GraphNode | tuple[GraphNode, str] = Field(default=True, description='Upsample restored faces for high-resolution AI-created images')
    background_enhance: bool | GraphNode | tuple[GraphNode, str] = Field(default=True, description='Enhance background image with Real-ESRGAN')
    codeformer_fidelity: float | GraphNode | tuple[GraphNode, str] = Field(default=0.5, description='Balance the quality (lower number) and fidelity (higher number).')
    @classmethod
    def get_node_type(cls): return "replicate.image.enhance.CodeFormer"


from nodetool.nodes.replicate.image.enhance import Model

class Maxim(GraphNode):
    image: ImageRef | GraphNode | tuple[GraphNode, str] = Field(default=ImageRef(type='image', uri='', asset_id=None), description='Input image.')
    model: Model | GraphNode | tuple[GraphNode, str] = Field(default=None, description='Choose a model.')
    @classmethod
    def get_node_type(cls): return "replicate.image.enhance.Maxim"



class OldPhotosRestoration(GraphNode):
    HR: bool | GraphNode | tuple[GraphNode, str] = Field(default=False, description='whether the input image is high-resolution')
    image: ImageRef | GraphNode | tuple[GraphNode, str] = Field(default=ImageRef(type='image', uri='', asset_id=None), description='input image.')
    with_scratch: bool | GraphNode | tuple[GraphNode, str] = Field(default=False, description='whether the input image is scratched')
    @classmethod
    def get_node_type(cls): return "replicate.image.enhance.OldPhotosRestoration"





from pydantic import BaseModel, Field
import nodetool.metadata.types
from nodetool.metadata.types import *
from nodetool.dsl.graph import GraphNode


class Audio(GraphNode):
    value: AudioRef | GraphNode | tuple[GraphNode, str] = Field(default=AudioRef(type='audio', uri='', asset_id=None, temp_id=None), description=None)
    @classmethod
    def get_node_type(cls): return "nodetool.constant.Audio"



class Bool(GraphNode):
    value: bool | GraphNode | tuple[GraphNode, str] = Field(default=False, description=None)
    @classmethod
    def get_node_type(cls): return "nodetool.constant.Bool"



class Constant(GraphNode):
    @classmethod
    def get_node_type(cls): return "nodetool.constant.Constant"



class DataFrame(GraphNode):
    value: DataframeRef | GraphNode | tuple[GraphNode, str] = Field(default=DataframeRef(type='dataframe', uri='', asset_id=None, temp_id=None, columns=None, data=None), description=None)
    @classmethod
    def get_node_type(cls): return "nodetool.constant.DataFrame"



class Dict(GraphNode):
    value: dict[str, Any] | GraphNode | tuple[GraphNode, str] = Field(default={}, description=None)
    @classmethod
    def get_node_type(cls): return "nodetool.constant.Dict"



class Float(GraphNode):
    value: float | GraphNode | tuple[GraphNode, str] = Field(default=0.0, description=None)
    @classmethod
    def get_node_type(cls): return "nodetool.constant.Float"



class Image(GraphNode):
    value: ImageRef | GraphNode | tuple[GraphNode, str] = Field(default=ImageRef(type='image', uri='', asset_id=None, temp_id=None), description=None)
    @classmethod
    def get_node_type(cls): return "nodetool.constant.Image"



class List(GraphNode):
    value: list[Any] | GraphNode | tuple[GraphNode, str] = Field(default=[], description=None)
    @classmethod
    def get_node_type(cls): return "nodetool.constant.List"



class String(GraphNode):
    value: str | GraphNode | tuple[GraphNode, str] = Field(default='', description=None)
    @classmethod
    def get_node_type(cls): return "nodetool.constant.String"



class Text(GraphNode):
    value: TextRef | GraphNode | tuple[GraphNode, str] = Field(default=TextRef(type='text', uri='', asset_id=None, temp_id=None), description=None)
    @classmethod
    def get_node_type(cls): return "nodetool.constant.Text"



class Video(GraphNode):
    value: VideoRef | GraphNode | tuple[GraphNode, str] = Field(default=VideoRef(type='video', uri='', asset_id=None, temp_id=None, duration=None, format=None), description=None)
    @classmethod
    def get_node_type(cls): return "nodetool.constant.Video"





#!/usr/bin/env python

from nodetool.common.environment import Environment
from nodetool.api.utils import current_user
from nodetool.models.user import User
from fastapi import APIRouter, Depends, HTTPException, Response

log = Environment.get_logger()
router = APIRouter(prefix="/api/models", tags=["models"])


@router.get("/{folder}")
async def index(folder: str, user: User = Depends(current_user)) -> list[str]:
    try:
        return Environment.get_model_files(folder)
    except Exception as e:
        raise HTTPException(status_code=404, detail="Error getting model files")



from pydantic import BaseModel


from typing import Any, Literal

from nodetool.api.types.prediction import Prediction


class NodeUpdate(BaseModel):
    type: Literal["node_update"] = "node_update"
    node_id: str
    node_name: str
    status: str
    error: str | None = None
    logs: str | None = None
    result: dict[str, Any] | None = None
    properties: dict[str, Any] | None = None
    started_at: str | None = None
    completed_at: str | None = None


class NodeProgress(BaseModel):
    type: Literal["node_progress"] = "node_progress"
    node_id: str
    progress: int
    total: int


class WorkflowUpdate(BaseModel):
    type: Literal["workflow_update"] = "workflow_update"
    result: dict[str, Any]


class Error(BaseModel):
    type: Literal["error"] = "error"
    error: str


ProcessingMessage = NodeUpdate | NodeProgress | WorkflowUpdate | Error | Prediction



from pydantic import BaseModel, Field
import nodetool.metadata.types
from nodetool.metadata.types import *
from nodetool.dsl.graph import GraphNode


class Chunk(GraphNode):
    text: str | nodetool.metadata.types.TextRef | GraphNode | tuple[GraphNode, str] = Field(default='', description=None)
    length: int | GraphNode | tuple[GraphNode, str] = Field(default=100, description=None)
    overlap: int | GraphNode | tuple[GraphNode, str] = Field(default=0, description=None)
    separator: str | None | GraphNode | tuple[GraphNode, str] = Field(default=None, description=None)
    @classmethod
    def get_node_type(cls): return "nodetool.text.extract.Chunk"



class Embedding(GraphNode):
    input: str | nodetool.metadata.types.TextRef | GraphNode | tuple[GraphNode, str] = Field(default='', description=None)
    model: LlamaModel | GraphNode | tuple[GraphNode, str] = Field(default=LlamaModel(type='llama_model', name='', repo_id='', filename='', local_path=None), description=None)
    n_gpu_layers: int | GraphNode | tuple[GraphNode, str] = Field(default=0, description='The number of layers on the GPU')
    chunk_size: int | GraphNode | tuple[GraphNode, str] = Field(default=4096, description='The size of the chunks to split the input into')
    @classmethod
    def get_node_type(cls): return "nodetool.text.extract.Embedding"



class Extract(GraphNode):
    text: str | nodetool.metadata.types.TextRef | GraphNode | tuple[GraphNode, str] = Field(default='', description=None)
    start: int | GraphNode | tuple[GraphNode, str] = Field(default=0, description=None)
    end: int | GraphNode | tuple[GraphNode, str] = Field(default=0, description=None)
    @classmethod
    def get_node_type(cls): return "nodetool.text.extract.Extract"



class ExtractJSON(GraphNode):
    text: str | nodetool.metadata.types.TextRef | GraphNode | tuple[GraphNode, str] = Field(default=PydanticUndefined, description=None)
    json_path: str | GraphNode | tuple[GraphNode, str] = Field(default='$.*', description=None)
    find_all: bool | GraphNode | tuple[GraphNode, str] = Field(default=False, description=None)
    @classmethod
    def get_node_type(cls): return "nodetool.text.extract.ExtractJSON"



class ExtractRegex(GraphNode):
    text: str | nodetool.metadata.types.TextRef | GraphNode | tuple[GraphNode, str] = Field(default='', description=None)
    regex: str | GraphNode | tuple[GraphNode, str] = Field(default='', description=None)
    dotall: bool | GraphNode | tuple[GraphNode, str] = Field(default=False, description=None)
    ignorecase: bool | GraphNode | tuple[GraphNode, str] = Field(default=False, description=None)
    multiline: bool | GraphNode | tuple[GraphNode, str] = Field(default=False, description=None)
    @classmethod
    def get_node_type(cls): return "nodetool.text.extract.ExtractRegex"



class FindAllRegex(GraphNode):
    text: str | nodetool.metadata.types.TextRef | GraphNode | tuple[GraphNode, str] = Field(default='', description=None)
    regex: str | GraphNode | tuple[GraphNode, str] = Field(default='', description=None)
    dotall: bool | GraphNode | tuple[GraphNode, str] = Field(default=False, description=None)
    ignorecase: bool | GraphNode | tuple[GraphNode, str] = Field(default=False, description=None)
    multiline: bool | GraphNode | tuple[GraphNode, str] = Field(default=False, description=None)
    @classmethod
    def get_node_type(cls): return "nodetool.text.extract.FindAllRegex"



class ParseJSON(GraphNode):
    text: str | nodetool.metadata.types.TextRef | GraphNode | tuple[GraphNode, str] = Field(default='', description=None)
    @classmethod
    def get_node_type(cls): return "nodetool.text.extract.ParseJSON"





from pydantic import BaseModel, Field
import nodetool.metadata.types
from nodetool.metadata.types import *
from nodetool.dsl.graph import GraphNode


class AdaptiveContrast(GraphNode):
    image: ImageRef | GraphNode | tuple[GraphNode, str] = Field(default=ImageRef(type='image', uri='', asset_id=None, temp_id=None), description='The image to adjust the contrast for.')
    clip_limit: float | GraphNode | tuple[GraphNode, str] = Field(default=2.0, description='Clip limit for adaptive contrast.')
    grid_size: int | GraphNode | tuple[GraphNode, str] = Field(default=8, description='Grid size for adaptive contrast.')
    @classmethod
    def get_node_type(cls): return "nodetool.image.enhance.AdaptiveContrast"



class AutoContrast(GraphNode):
    image: ImageRef | GraphNode | tuple[GraphNode, str] = Field(default=ImageRef(type='image', uri='', asset_id=None, temp_id=None), description='The image to adjust the contrast for.')
    cutoff: int | GraphNode | tuple[GraphNode, str] = Field(default=0, description='Cutoff for autocontrast.')
    @classmethod
    def get_node_type(cls): return "nodetool.image.enhance.AutoContrast"



class Brightness(GraphNode):
    image: ImageRef | GraphNode | tuple[GraphNode, str] = Field(default=ImageRef(type='image', uri='', asset_id=None, temp_id=None), description='The image to adjust the brightness for.')
    factor: float | int | GraphNode | tuple[GraphNode, str] = Field(default=1.0, description='Factor to adjust the brightness. 1.0 means no change.')
    @classmethod
    def get_node_type(cls): return "nodetool.image.enhance.Brightness"



class Color(GraphNode):
    image: ImageRef | GraphNode | tuple[GraphNode, str] = Field(default=ImageRef(type='image', uri='', asset_id=None, temp_id=None), description='The image to adjust the brightness for.')
    factor: float | GraphNode | tuple[GraphNode, str] = Field(default=1.0, description='Factor to adjust the contrast. 1.0 means no change.')
    @classmethod
    def get_node_type(cls): return "nodetool.image.enhance.Color"



class Contrast(GraphNode):
    image: ImageRef | GraphNode | tuple[GraphNode, str] = Field(default=ImageRef(type='image', uri='', asset_id=None, temp_id=None), description='The image to adjust the brightness for.')
    factor: float | GraphNode | tuple[GraphNode, str] = Field(default=1.0, description='Factor to adjust the contrast. 1.0 means no change.')
    @classmethod
    def get_node_type(cls): return "nodetool.image.enhance.Contrast"



class Detail(GraphNode):
    image: ImageRef | GraphNode | tuple[GraphNode, str] = Field(default=ImageRef(type='image', uri='', asset_id=None, temp_id=None), description='The image to detail.')
    @classmethod
    def get_node_type(cls): return "nodetool.image.enhance.Detail"



class EdgeEnhance(GraphNode):
    image: ImageRef | GraphNode | tuple[GraphNode, str] = Field(default=ImageRef(type='image', uri='', asset_id=None, temp_id=None), description='The image to edge enhance.')
    @classmethod
    def get_node_type(cls): return "nodetool.image.enhance.EdgeEnhance"



class Equalize(GraphNode):
    image: ImageRef | GraphNode | tuple[GraphNode, str] = Field(default=ImageRef(type='image', uri='', asset_id=None, temp_id=None), description='The image to equalize.')
    @classmethod
    def get_node_type(cls): return "nodetool.image.enhance.Equalize"



class RankFilter(GraphNode):
    image: ImageRef | GraphNode | tuple[GraphNode, str] = Field(default=ImageRef(type='image', uri='', asset_id=None, temp_id=None), description='The image to rank filter.')
    size: int | GraphNode | tuple[GraphNode, str] = Field(default=3, description='Rank filter size.')
    rank: int | GraphNode | tuple[GraphNode, str] = Field(default=3, description='Rank filter rank.')
    @classmethod
    def get_node_type(cls): return "nodetool.image.enhance.RankFilter"



class Sharpen(GraphNode):
    image: ImageRef | GraphNode | tuple[GraphNode, str] = Field(default=ImageRef(type='image', uri='', asset_id=None, temp_id=None), description='The image to sharpen.')
    @classmethod
    def get_node_type(cls): return "nodetool.image.enhance.Sharpen"



class Sharpness(GraphNode):
    image: ImageRef | GraphNode | tuple[GraphNode, str] = Field(default=ImageRef(type='image', uri='', asset_id=None, temp_id=None), description='The image to adjust the brightness for.')
    factor: float | GraphNode | tuple[GraphNode, str] = Field(default=1.0, description='Factor to adjust the contrast. 1.0 means no change.')
    @classmethod
    def get_node_type(cls): return "nodetool.image.enhance.Sharpness"



class UnsharpMask(GraphNode):
    image: ImageRef | GraphNode | tuple[GraphNode, str] = Field(default=ImageRef(type='image', uri='', asset_id=None, temp_id=None), description='The image to unsharp mask.')
    radius: int | GraphNode | tuple[GraphNode, str] = Field(default=2, description='Unsharp mask radius.')
    percent: int | GraphNode | tuple[GraphNode, str] = Field(default=150, description='Unsharp mask percent.')
    threshold: int | GraphNode | tuple[GraphNode, str] = Field(default=3, description='Unsharp mask threshold.')
    @classmethod
    def get_node_type(cls): return "nodetool.image.enhance.UnsharpMask"





import os
import random

import cv2
import numpy as np
import torch
from pathlib import Path
import warnings
from huggingface_hub import hf_hub_download

HF_MODEL_NAME = "lllyasviel/Annotators"
DWPOSE_MODEL_NAME = "yzd-v/DWPose"
ANIFACESEG_MODEL_NAME = "bdsqlsz/qinglong_controlnet-lllite"
DENSEPOSE_MODEL_NAME = "LayerNorm/DensePose-TorchScript-with-hint-image"
MESH_GRAPHORMER_MODEL_NAME = "hr16/ControlNet-HandRefiner-pruned"
SAM_MODEL_NAME = "dhkim2810/MobileSAM"

annotator_ckpts_path = os.path.join(Path(__file__).parents[2], 'ckpts')
USE_SYMLINKS = False

try:
    annotator_ckpts_path = os.environ['AUX_ANNOTATOR_CKPTS_PATH']
except:
    warnings.warn("Custom pressesor model path not set successfully.")
    pass

try:
    USE_SYMLINKS = eval(os.environ['AUX_USE_SYMLINKS'])
except:
    warnings.warn("USE_SYMLINKS not set successfully. Using default value: False to download models.")
    pass

# fix SSL: CERTIFICATE_VERIFY_FAILED issue with pytorch download https://github.com/pytorch/pytorch/issues/33288
try:
    from torch.hub import load_state_dict_from_url
    test_url = "https://download.pytorch.org/models/mobilenet_v2-b0353104.pth"
    load_state_dict_from_url(test_url, progress=False)
except:
    import ssl
    ssl._create_default_https_context = ssl._create_unverified_context

here = Path(__file__).parent.resolve()

def HWC3(x):
    assert x.dtype == np.uint8
    if x.ndim == 2:
        x = x[:, :, None]
    assert x.ndim == 3
    H, W, C = x.shape
    assert C == 1 or C == 3 or C == 4
    if C == 3:
        return x
    if C == 1:
        return np.concatenate([x, x, x], axis=2)
    if C == 4:
        color = x[:, :, 0:3].astype(np.float32)
        alpha = x[:, :, 3:4].astype(np.float32) / 255.0
        y = color * alpha + 255.0 * (1.0 - alpha)
        y = y.clip(0, 255).astype(np.uint8)
        return y


def make_noise_disk(H, W, C, F):
    noise = np.random.uniform(low=0, high=1, size=((H // F) + 2, (W // F) + 2, C))
    noise = cv2.resize(noise, (W + 2 * F, H + 2 * F), interpolation=cv2.INTER_CUBIC)
    noise = noise[F: F + H, F: F + W]
    noise -= np.min(noise)
    noise /= np.max(noise)
    if C == 1:
        noise = noise[:, :, None]
    return noise


def nms(x, t, s):
    x = cv2.GaussianBlur(x.astype(np.float32), (0, 0), s)

    f1 = np.array([[0, 0, 0], [1, 1, 1], [0, 0, 0]], dtype=np.uint8)
    f2 = np.array([[0, 1, 0], [0, 1, 0], [0, 1, 0]], dtype=np.uint8)
    f3 = np.array([[1, 0, 0], [0, 1, 0], [0, 0, 1]], dtype=np.uint8)
    f4 = np.array([[0, 0, 1], [0, 1, 0], [1, 0, 0]], dtype=np.uint8)

    y = np.zeros_like(x)

    for f in [f1, f2, f3, f4]:
        np.putmask(y, cv2.dilate(x, kernel=f) == x, x)

    z = np.zeros_like(y, dtype=np.uint8)
    z[y > t] = 255
    return z

def min_max_norm(x):
    x -= np.min(x)
    x /= np.maximum(np.max(x), 1e-5)
    return x


def safe_step(x, step=2):
    y = x.astype(np.float32) * float(step + 1)
    y = y.astype(np.int32).astype(np.float32) / float(step)
    return y


def img2mask(img, H, W, low=10, high=90):
    assert img.ndim == 3 or img.ndim == 2
    assert img.dtype == np.uint8

    if img.ndim == 3:
        y = img[:, :, random.randrange(0, img.shape[2])]
    else:
        y = img

    y = cv2.resize(y, (W, H), interpolation=cv2.INTER_CUBIC)

    if random.uniform(0, 1) < 0.5:
        y = 255 - y

    return y < np.percentile(y, random.randrange(low, high))

def safer_memory(x):
    # Fix many MAC/AMD problems
    return np.ascontiguousarray(x.copy()).copy()

UPSCALE_METHODS = ["INTER_NEAREST", "INTER_LINEAR", "INTER_AREA", "INTER_CUBIC", "INTER_LANCZOS4"]
def get_upscale_method(method_str):
    assert method_str in UPSCALE_METHODS, f"Method {method_str} not found in {UPSCALE_METHODS}"
    return getattr(cv2, method_str)

def pad64(x):
    return int(np.ceil(float(x) / 64.0) * 64 - x)

#https://github.com/Mikubill/sd-webui-controlnet/blob/main/scripts/processor.py#L17
#Added upscale_method param
def resize_image_with_pad(input_image, resolution, upscale_method = "", skip_hwc3=False):
    if skip_hwc3:
        img = input_image
    else:
        img = HWC3(input_image)
    H_raw, W_raw, _ = img.shape
    k = float(resolution) / float(min(H_raw, W_raw))
    H_target = int(np.round(float(H_raw) * k))
    W_target = int(np.round(float(W_raw) * k))
    img = cv2.resize(img, (W_target, H_target), interpolation=get_upscale_method(upscale_method) if k > 1 else cv2.INTER_AREA)
    H_pad, W_pad = pad64(H_target), pad64(W_target)
    img_padded = np.pad(img, [[0, H_pad], [0, W_pad], [0, 0]], mode='edge')

    def remove_pad(x):
        return safer_memory(x[:H_target, :W_target, ...])

    return safer_memory(img_padded), remove_pad
    
def common_input_validate(input_image, output_type, **kwargs):
    if "img" in kwargs:
            warnings.warn("img is deprecated, please use `input_image=...` instead.", DeprecationWarning)
            input_image = kwargs.pop("img")
    
    if "return_pil" in kwargs:
            warnings.warn("return_pil is deprecated. Use output_type instead.", DeprecationWarning)
            output_type = "pil" if kwargs["return_pil"] else "np"
    
    if type(output_type) is bool:
        warnings.warn("Passing `True` or `False` to `output_type` is deprecated and will raise an error in future versions")
        if output_type:
            output_type = "pil"

    if input_image is None:
        raise ValueError("input_image must be defined.")

    if not isinstance(input_image, np.ndarray):
        input_image = np.array(input_image, dtype=np.uint8)
        output_type = output_type or "pil"
    else:
        output_type = output_type or "np"
    
    return (input_image, output_type)

def torch_gc():
    if torch.cuda.is_available():
        torch.cuda.empty_cache()
        torch.cuda.ipc_collect()


def ade_palette():
    """ADE20K palette that maps each class to RGB values."""
    return [[120, 120, 120], [180, 120, 120], [6, 230, 230], [80, 50, 50],
            [4, 200, 3], [120, 120, 80], [140, 140, 140], [204, 5, 255],
            [230, 230, 230], [4, 250, 7], [224, 5, 255], [235, 255, 7],
            [150, 5, 61], [120, 120, 70], [8, 255, 51], [255, 6, 82],
            [143, 255, 140], [204, 255, 4], [255, 51, 7], [204, 70, 3],
            [0, 102, 200], [61, 230, 250], [255, 6, 51], [11, 102, 255],
            [255, 7, 71], [255, 9, 224], [9, 7, 230], [220, 220, 220],
            [255, 9, 92], [112, 9, 255], [8, 255, 214], [7, 255, 224],
            [255, 184, 6], [10, 255, 71], [255, 41, 10], [7, 255, 255],
            [224, 255, 8], [102, 8, 255], [255, 61, 6], [255, 194, 7],
            [255, 122, 8], [0, 255, 20], [255, 8, 41], [255, 5, 153],
            [6, 51, 255], [235, 12, 255], [160, 150, 20], [0, 163, 255],
            [140, 140, 140], [250, 10, 15], [20, 255, 0], [31, 255, 0],
            [255, 31, 0], [255, 224, 0], [153, 255, 0], [0, 0, 255],
            [255, 71, 0], [0, 235, 255], [0, 173, 255], [31, 0, 255],
            [11, 200, 200], [255, 82, 0], [0, 255, 245], [0, 61, 255],
            [0, 255, 112], [0, 255, 133], [255, 0, 0], [255, 163, 0],
            [255, 102, 0], [194, 255, 0], [0, 143, 255], [51, 255, 0],
            [0, 82, 255], [0, 255, 41], [0, 255, 173], [10, 0, 255],
            [173, 255, 0], [0, 255, 153], [255, 92, 0], [255, 0, 255],
            [255, 0, 245], [255, 0, 102], [255, 173, 0], [255, 0, 20],
            [255, 184, 184], [0, 31, 255], [0, 255, 61], [0, 71, 255],
            [255, 0, 204], [0, 255, 194], [0, 255, 82], [0, 10, 255],
            [0, 112, 255], [51, 0, 255], [0, 194, 255], [0, 122, 255],
            [0, 255, 163], [255, 153, 0], [0, 255, 10], [255, 112, 0],
            [143, 255, 0], [82, 0, 255], [163, 255, 0], [255, 235, 0],
            [8, 184, 170], [133, 0, 255], [0, 255, 92], [184, 0, 255],
            [255, 0, 31], [0, 184, 255], [0, 214, 255], [255, 0, 112],
            [92, 255, 0], [0, 224, 255], [112, 224, 255], [70, 184, 160],
            [163, 0, 255], [153, 0, 255], [71, 255, 0], [255, 0, 163],
            [255, 204, 0], [255, 0, 143], [0, 255, 235], [133, 255, 0],
            [255, 0, 235], [245, 0, 255], [255, 0, 122], [255, 245, 0],
            [10, 190, 212], [214, 255, 0], [0, 204, 255], [20, 0, 255],
            [255, 255, 0], [0, 153, 255], [0, 41, 255], [0, 255, 204],
            [41, 0, 255], [41, 255, 0], [173, 0, 255], [0, 245, 255],
            [71, 0, 255], [122, 0, 255], [0, 255, 184], [0, 92, 255],
            [184, 255, 0], [0, 133, 255], [255, 214, 0], [25, 194, 194],
            [102, 255, 0], [92, 0, 255]]

def custom_hf_download(pretrained_model_or_path, filename, cache_dir=annotator_ckpts_path, subfolder='', use_symlinks=USE_SYMLINKS):
    # Fix for Windows, added by @georgi
    path = os.path.join(*pretrained_model_or_path.split('/'))
    local_dir = os.path.join(cache_dir, path)
    model_path = os.path.join(local_dir, *subfolder.split('/'), filename)
    
    if not os.path.exists(model_path):
        print(f"Failed to find {model_path}.\n Downloading from huggingface.co")
        if use_symlinks:
            cache_dir_d = os.getenv("HUGGINGFACE_HUB_CACHE")
            if cache_dir_d is None:
                import platform
                if platform.system() == "Windows":
                    cache_dir_d = os.path.join(os.getenv("USERPROFILE"), ".cache", "huggingface", "hub")
                else:
                    cache_dir_d = os.path.join(os.getenv("HOME"), ".cache", "huggingface", "hub")
            try:
                # test_link
                if not os.path.exists(cache_dir_d):
                    os.makedirs(cache_dir_d)
                open(os.path.join(cache_dir_d, f"linktest_{filename}.txt"), "w")
                os.link(os.path.join(cache_dir_d, f"linktest_{filename}.txt"), os.path.join(cache_dir, f"linktest_{filename}.txt"))
                os.remove(os.path.join(cache_dir, f"linktest_{filename}.txt"))
                os.remove(os.path.join(cache_dir_d, f"linktest_{filename}.txt"))
                print("Using symlinks to download models. \n",\
                      "Make sure you have enough space on your cache folder. \n",\
                      "And do not purge the cache folder after downloading.\n",\
                      "Otherwise, you will have to re-download the models every time you run the script.\n",\
                      "You can use USE_SYMLINKS: False in config.yaml to avoid this behavior.")
            except:
                print("Maybe not able to create symlink. Disable using symlinks.")
                use_symlinks = False
                cache_dir_d = os.path.join(cache_dir, pretrained_model_or_path, "cache")
        else:
            cache_dir_d = os.path.join(cache_dir, pretrained_model_or_path, "cache")

        model_path = hf_hub_download(repo_id=pretrained_model_or_path,
            cache_dir=cache_dir_d,
            local_dir=local_dir,
            subfolder=subfolder,
            filename=filename,
            local_dir_use_symlinks=use_symlinks,
            resume_download=True,
            etag_timeout=100
        )
        if not use_symlinks:
            try:
                import shutil
                shutil.rmtree(cache_dir_d)
            except Exception as e :
                print(e)
    return model_path


from pydantic import BaseModel, Field
import nodetool.metadata.types
from nodetool.metadata.types import *
from nodetool.dsl.graph import GraphNode


class Blur(GraphNode):
    image: ImageRef | GraphNode | tuple[GraphNode, str] = Field(default=ImageRef(type='image', uri='', asset_id=None, temp_id=None), description='The image to blur.')
    radius: float | GraphNode | tuple[GraphNode, str] = Field(default=2.0, description='Blur radius.')
    @classmethod
    def get_node_type(cls): return "nodetool.image.transform.Blur"



class Canny(GraphNode):
    image: ImageRef | GraphNode | tuple[GraphNode, str] = Field(default=ImageRef(type='image', uri='', asset_id=None, temp_id=None), description='The image to canny.')
    low_threshold: int | GraphNode | tuple[GraphNode, str] = Field(default=100, description='Low threshold.')
    high_threshold: int | GraphNode | tuple[GraphNode, str] = Field(default=200, description='High threshold.')
    @classmethod
    def get_node_type(cls): return "nodetool.image.transform.Canny"



class Contour(GraphNode):
    image: ImageRef | GraphNode | tuple[GraphNode, str] = Field(default=ImageRef(type='image', uri='', asset_id=None, temp_id=None), description='The image to contour.')
    @classmethod
    def get_node_type(cls): return "nodetool.image.transform.Contour"



class ConvertToGrayscale(GraphNode):
    image: ImageRef | GraphNode | tuple[GraphNode, str] = Field(default=ImageRef(type='image', uri='', asset_id=None, temp_id=None), description='The image to convert.')
    @classmethod
    def get_node_type(cls): return "nodetool.image.transform.ConvertToGrayscale"



class Crop(GraphNode):
    image: ImageRef | GraphNode | tuple[GraphNode, str] = Field(default=ImageRef(type='image', uri='', asset_id=None, temp_id=None), description='The image to crop.')
    left: int | GraphNode | tuple[GraphNode, str] = Field(default=0, description='The left coordinate.')
    top: int | GraphNode | tuple[GraphNode, str] = Field(default=0, description='The top coordinate.')
    right: int | GraphNode | tuple[GraphNode, str] = Field(default=512, description='The right coordinate.')
    bottom: int | GraphNode | tuple[GraphNode, str] = Field(default=512, description='The bottom coordinate.')
    @classmethod
    def get_node_type(cls): return "nodetool.image.transform.Crop"



class Emboss(GraphNode):
    image: ImageRef | GraphNode | tuple[GraphNode, str] = Field(default=ImageRef(type='image', uri='', asset_id=None, temp_id=None), description='The image to emboss.')
    @classmethod
    def get_node_type(cls): return "nodetool.image.transform.Emboss"



class Expand(GraphNode):
    image: ImageRef | GraphNode | tuple[GraphNode, str] = Field(default=ImageRef(type='image', uri='', asset_id=None, temp_id=None), description='The image to expand.')
    border: int | GraphNode | tuple[GraphNode, str] = Field(default=0, description='Border size.')
    fill: int | GraphNode | tuple[GraphNode, str] = Field(default=0, description='Fill color.')
    @classmethod
    def get_node_type(cls): return "nodetool.image.transform.Expand"



class FindEdges(GraphNode):
    image: ImageRef | GraphNode | tuple[GraphNode, str] = Field(default=ImageRef(type='image', uri='', asset_id=None, temp_id=None), description='The image to find edges.')
    @classmethod
    def get_node_type(cls): return "nodetool.image.transform.FindEdges"



class Fit(GraphNode):
    image: ImageRef | GraphNode | tuple[GraphNode, str] = Field(default=ImageRef(type='image', uri='', asset_id=None, temp_id=None), description='The image to fit.')
    width: int | GraphNode | tuple[GraphNode, str] = Field(default=512, description='Width to fit to.')
    height: int | GraphNode | tuple[GraphNode, str] = Field(default=512, description='Height to fit to.')
    @classmethod
    def get_node_type(cls): return "nodetool.image.transform.Fit"


from nodetool.nodes.nodetool.image.transform import ChannelEnum

class GetChannel(GraphNode):
    image: ImageRef | GraphNode | tuple[GraphNode, str] = Field(default=ImageRef(type='image', uri='', asset_id=None, temp_id=None), description='The image to get the channel from.')
    channel: ChannelEnum | GraphNode | tuple[GraphNode, str] = Field(default=ChannelEnum('R'), description=None)
    @classmethod
    def get_node_type(cls): return "nodetool.image.transform.GetChannel"



class Invert(GraphNode):
    image: ImageRef | GraphNode | tuple[GraphNode, str] = Field(default=ImageRef(type='image', uri='', asset_id=None, temp_id=None), description='The image to adjust the brightness for.')
    @classmethod
    def get_node_type(cls): return "nodetool.image.transform.Invert"



class Posterize(GraphNode):
    image: ImageRef | GraphNode | tuple[GraphNode, str] = Field(default=ImageRef(type='image', uri='', asset_id=None, temp_id=None), description='The image to posterize.')
    bits: int | GraphNode | tuple[GraphNode, str] = Field(default=4, description='Number of bits to posterize to.')
    @classmethod
    def get_node_type(cls): return "nodetool.image.transform.Posterize"



class Resize(GraphNode):
    image: ImageRef | GraphNode | tuple[GraphNode, str] = Field(default=ImageRef(type='image', uri='', asset_id=None, temp_id=None), description='The image to resize.')
    width: int | GraphNode | tuple[GraphNode, str] = Field(default=512, description='The target width.')
    height: int | GraphNode | tuple[GraphNode, str] = Field(default=512, description='The target height.')
    @classmethod
    def get_node_type(cls): return "nodetool.image.transform.Resize"



class Scale(GraphNode):
    image: ImageRef | GraphNode | tuple[GraphNode, str] = Field(default=ImageRef(type='image', uri='', asset_id=None, temp_id=None), description='The image to scale.')
    scale: float | GraphNode | tuple[GraphNode, str] = Field(default=1.0, description='The scale factor.')
    @classmethod
    def get_node_type(cls): return "nodetool.image.transform.Scale"



class Smooth(GraphNode):
    image: ImageRef | GraphNode | tuple[GraphNode, str] = Field(default=ImageRef(type='image', uri='', asset_id=None, temp_id=None), description='The image to smooth.')
    @classmethod
    def get_node_type(cls): return "nodetool.image.transform.Smooth"



class Solarize(GraphNode):
    image: ImageRef | GraphNode | tuple[GraphNode, str] = Field(default=ImageRef(type='image', uri='', asset_id=None, temp_id=None), description='The image to solarize.')
    threshold: int | GraphNode | tuple[GraphNode, str] = Field(default=128, description='Threshold for solarization.')
    @classmethod
    def get_node_type(cls): return "nodetool.image.transform.Solarize"





__version__ = "1.0.0"

from .modeling_bert import (BertConfig, BertModel,
                       load_tf_weights_in_bert)

from .modeling_graphormer import Graphormer

from .e2e_body_network import Graphormer_Body_Network

from .e2e_hand_network import Graphormer_Hand_Network

CONFIG_NAME = "config.json"

from .modeling_utils import (WEIGHTS_NAME, TF_WEIGHTS_NAME,
                          PretrainedConfig, PreTrainedModel, prune_layer, Conv1D)

from .file_utils import (PYTORCH_PRETRAINED_BERT_CACHE)



from fastapi import (
    HTTPException,
    Header,
    status,
)
from pydantic import BaseModel
from typing import Optional
from fastapi import Cookie
from nodetool.common.environment import Environment

from nodetool.models.user import User

local_user: User | None = None

log = Environment.get_logger()


def get_local_user():
    """
    In local mode, we only need to create a single user.
    This single user has the ID 1.
    """
    global local_user
    if local_user is None:
        local_user = User.get(1)
        if local_user is None:
            local_user, _ = User.create(user_id="1", auth_token="local")
    return local_user


async def current_user(
    authorization: Optional[str] = Header(None),
    auth_cookie: Optional[str] = Cookie(None),
) -> User:
    if not Environment.use_remote_auth():
        return get_local_user()

    if authorization:
        token_split = authorization.split(" ")
        if len(token_split) != 2 or token_split[0].lower() != "bearer":
            raise HTTPException(status_code=status.HTTP_401_UNAUTHORIZED)

        user = User.find_by_auth_token(token_split[1])
    elif auth_cookie:
        user = User.find_by_auth_token(auth_cookie)
    else:
        user = None

    if not user:
        raise HTTPException(status_code=401, detail="User not authenticated")
    if user.verified_at is None:
        raise HTTPException(status_code=401, detail="User not verified")
    return user


async def abort(status_code: int, detail: Optional[str] = None) -> None:
    """
    Abort the current request with the given status code and detail.
    """
    raise HTTPException(status_code=status_code, detail=detail)



from pydantic import BaseModel, Field
import nodetool.metadata.types
from nodetool.metadata.types import *
from nodetool.dsl.graph import GraphNode


class ChromaNode(GraphNode):
    @classmethod
    def get_node_type(cls): return "nodetool.vector.Chroma"



class IndexFolder(GraphNode):
    folder: FolderRef | GraphNode | tuple[GraphNode, str] = Field(default=FolderRef(type='folder', uri='', asset_id=None, temp_id=None), description='The folder to index')
    @classmethod
    def get_node_type(cls): return "nodetool.vector.IndexFolder"



class NearestNeighbors(GraphNode):
    documents: list[Tensor] | GraphNode | tuple[GraphNode, str] = Field(default=[], description='The list of documents to search')
    query: Tensor | GraphNode | tuple[GraphNode, str] = Field(default=Tensor(type='tensor', value=[], dtype=None), description='The query to search for')
    n_neighbors: int | GraphNode | tuple[GraphNode, str] = Field(default=1, description='The number of neighbors to return')
    @classmethod
    def get_node_type(cls): return "nodetool.vector.NearestNeighbors"



class QueryImage(GraphNode):
    folder: FolderRef | GraphNode | tuple[GraphNode, str] = Field(default=FolderRef(type='folder', uri='', asset_id=None, temp_id=None), description='The folder to query')
    image: ImageRef | GraphNode | tuple[GraphNode, str] = Field(default=ImageRef(type='image', uri='', asset_id=None, temp_id=None), description='The image to query')
    n_results: int | GraphNode | tuple[GraphNode, str] = Field(default=1, description='The number of results to return')
    @classmethod
    def get_node_type(cls): return "nodetool.vector.QueryImage"



class QueryText(GraphNode):
    folder: FolderRef | GraphNode | tuple[GraphNode, str] = Field(default=FolderRef(type='folder', uri='', asset_id=None, temp_id=None), description='The folder to query')
    text: TextRef | GraphNode | tuple[GraphNode, str] = Field(default=TextRef(type='text', uri='', asset_id=None, temp_id=None), description='The text to query')
    n_results: int | GraphNode | tuple[GraphNode, str] = Field(default=1, description='The number of results to return')
    @classmethod
    def get_node_type(cls): return "nodetool.vector.QueryText"





from pydantic import BaseModel, Field
import nodetool.metadata.types
from nodetool.metadata.types import *
from nodetool.dsl.graph import GraphNode

from nodetool.nodes.nodetool.image.classify import Model

class Classify(GraphNode):
    model: Model | GraphNode | tuple[GraphNode, str] = Field(default=Model('microsoft/resnet-18'), description='The classification model to use')
    image: ImageRef | GraphNode | tuple[GraphNode, str] = Field(default=ImageRef(type='image', uri='', asset_id=None, temp_id=None), description='The image to classify')
    @classmethod
    def get_node_type(cls): return "nodetool.image.classify.Classify"





from enum import EnumMeta
import os
import inspect
import pkgutil
from types import GenericAlias, UnionType
from typing import get_type_hints, Union, Tuple
import importlib

from nodetool.metadata.utils import is_enum_type
from nodetool.workflows.base_node import BaseNode


"""
This script generates DSL modules based on all nodes in the source root module.
"""


def create_python_module_file(filename: str, content: str) -> None:
    """
    Create a Python module file with the given filename and content.

    Args:
      filename (str): The path and name of the file to be created.
      content (str): The content to be written to the file.

    Returns:
      None
    """
    os.makedirs(os.path.dirname(filename), exist_ok=True)
    with open(filename, "w") as file:
        file.write(content)


def type_to_string(field_type: type) -> str:
    """
    Converts a Python type to a string representation.

    Args:
      field_type (type): The Python type to convert.

    Returns:
      str: The string representation of the type.
    """
    if isinstance(field_type, UnionType):
        return str(field_type)
    if isinstance(field_type, dict):
        return "dict"
    if field_type == Union:
        return " | ".join(field_type.__args__)
    if isinstance(field_type, GenericAlias):
        args = [type_to_string(arg) for arg in field_type.__args__]
        if field_type.__origin__ == list:
            return f"list[{args[0]}]"
        elif field_type.__origin__ == dict:
            return f"dict[{args[0]}, {args[1]}]"
        elif field_type.__origin__ == set:
            return f"set[{args[0]}]"
        elif field_type.__origin__ == tuple:
            return f"tuple[{', '.join(args)}]"
    return field_type.__name__


def field_default(default_value):
    """
    Returns a string representation of the default value for a field.

    Args:
      default_value: The default value of the field.

    Returns:
      str: The string representation of the default value.

    """
    if default_value is None:
        return "None"
    if isinstance(type(default_value), EnumMeta):
        return f"{type(default_value).__name__}({repr(default_value.value)})"
    return repr(default_value)


def generate_class_source(node_cls: type[BaseNode]) -> str:
    """
    Generate the source code for a graph node class based on the provided node class.
    This class acts as a DSL wrapper for easier connection between nodes.

    Args:
      node_cls (type[BaseNode]): The node class to generate the source code for.

    Returns:
      str: The generated source code for the class.
    """

    imports = ""
    class_body = f"class {node_cls.__name__}(GraphNode):\n"
    fields = node_cls.inherited_fields()
    node_type = node_cls.get_node_type()

    for field_name, field_type in node_cls.field_types().items():
        if not field_name in fields:
            continue
        if is_enum_type(field_type):
            imports += f"from {field_type.__module__} import {field_type.__name__}\n"  # type: ignore
        field = fields[field_name]
        new_field_type = f"{type_to_string(field_type)} | GraphNode | tuple[GraphNode, str]"  # type: ignore
        field_def = f"Field(default={field_default(field.default)}, description={repr(field.description)})"
        class_body += f"    {field_name}: {new_field_type} = {field_def}\n"

    class_body += "    @classmethod\n"
    class_body += f'    def get_node_type(cls): return "{node_type}"\n'

    return imports + "\n" + class_body


def create_dsl_modules(source_root: str, target_root: str):
    """
    Generate DSL modules based on the source root and target root.

    Args:
      source_root (str): The source root module name.
      target_root (str): The target root module name.
    """
    source_root_module = importlib.import_module(source_root)
    target_root_module = importlib.import_module(target_root)
    target_root_path = str(target_root_module.__path__[0])

    for _, module_name, _ in pkgutil.walk_packages(
        source_root_module.__path__, prefix=source_root_module.__name__ + "."
    ):
        print(module_name)
        target_module_name: str = module_name.replace(source_root, target_root)
        module = importlib.import_module(module_name)
        source_code = ""
        for _, obj in inspect.getmembers(module, inspect.isclass):
            if issubclass(obj, BaseNode) and obj is not BaseNode and obj.is_visible():
                source_code += generate_class_source(obj)
                source_code += "\n\n"

        if source_code == "":
            continue

        target_module_path = target_module_name.replace(
            target_root_module.__name__ + ".", ""
        ).replace(".", "/")

        target_path = os.path.join(target_root_path, target_module_path) + ".py"
        source_code = (
            "from pydantic import BaseModel, Field\n"
            "import nodetool.metadata.types\n"
            "from nodetool.metadata.types import *\n"
            "from nodetool.dsl.graph import GraphNode\n\n"
        ) + source_code
        create_python_module_file(target_path, source_code)


if __name__ == "__main__":
    create_dsl_modules("nodetool.nodes", "nodetool.dsl")



from enum import Enum
from pydantic import Field, validator
from nodetool.metadata.types import LlamaModel
from nodetool.workflows.processing_context import ProcessingContext
from nodetool.workflows.base_node import BaseNode


class GPT2(BaseNode):
    """
    GPT-2 is a transformer-based language model. This node uses the GPT-2 model to generate text based on a prompt.

    # Applications
    - Generating text based on a prompt.
    - Generating text for chatbots.
    - Generating text for creative writing.
    """

    prompt: str = Field(default="", description="Prompt to send to the model.")
    max_tokens: int = Field(
        default=128,
        ge=1,
        le=1024,
        description="Maximum number of tokens to generate. A word is generally 2-3 tokens (minimum: 1)",
    )
    temperature: float = Field(
        default=0.7,
        ge=0.0,
        le=1.0,
        description="The temperature to use for the model.",
    )
    top_k: int = Field(
        default=50,
        ge=1,
        le=100,
        description="The number of highest probability tokens to keep for top-k sampling.",
    )
    top_p: float = Field(
        default=0.95,
        ge=0.0,
        le=1.0,
        description="The cumulative probability cutoff for nucleus/top-p sampling.",
    )

    async def process(self, context: ProcessingContext) -> str:
        from transformers import pipeline

        pipe = pipeline(task="text-generation", model="gpt2")
        res = pipe(
            self.prompt,
            max_new_tokens=self.max_tokens,
            do_sample=True,
            temperature=self.temperature,
            top_k=self.top_k,
            top_p=self.top_p,
            return_full_text=False,
        )
        return res[0]["generated_text"]  # type: ignore


class LlamaCpp(BaseNode):
    """
    Run Llama models.
    """

    model: LlamaModel = Field(
        default=LlamaModel(), description="The Llama model to use."
    )
    prompt: str = Field(default="", description="Prompt to send to the model.")
    system_prompt: str = Field(
        default="You are an assistant.",
        description="System prompt to send to the model.",
    )
    max_tokens: int = Field(
        default=128,
        ge=1,
        le=1024,
        description="Maximum number of tokens to generate. A word is generally 2-3 tokens (minimum: 1)",
    )
    temperature: float = Field(
        default=0.7,
        ge=0.0,
        le=1.0,
        description="The temperature to use for the model.",
    )
    n_gpu_layers: int = Field(
        default=0,
        ge=-1,
        le=100,
        description="Number of layers to offload to GPU (-ngl). If -1, all layers are offloaded.",
    )
    top_k: int = Field(
        default=50,
        ge=1,
        le=100,
        description="The number of highest probability tokens to keep for top-k sampling.",
    )
    top_p: float = Field(
        default=0.95,
        ge=0.0,
        le=1.0,
        description="The cumulative probability cutoff for nucleus/top-p sampling.",
    )
    grammar: str = Field(
        default="",
        description="The grammar to use for the model. If empty, no grammar is used.",
    )
    is_json_schema: bool = Field(
        default=False,
        description="Whether the grammar is a JSON schema. If true, the grammar is used as a JSON schema.",
    )

    @validator("model", pre=True)
    def validate_model(cls, v):
        if isinstance(v, str):
            v = LlamaModel(name=v)
        if isinstance(v, dict):
            v = LlamaModel(**v)
        if v.name == "":
            raise ValueError("The model cannot be empty.")
        return v

    async def process(self, context: ProcessingContext) -> str:
        from llama_cpp import LlamaGrammar

        if self.grammar != "":
            if self.is_json_schema:
                grammar = LlamaGrammar.from_json_schema(self.grammar)
            else:
                grammar = LlamaGrammar.from_string(self.grammar)
        else:
            grammar = None

        llm = context.load_llama_model(self.model, n_gpu_layers=self.n_gpu_layers)

        if llm is None:
            raise ValueError(f"Model {self.model.name} not found.")

        res = llm.create_chat_completion(
            messages=[
                {
                    "role": "system",
                    "content": self.system_prompt,
                },
                {"role": "user", "content": self.prompt},
            ],
            temperature=self.temperature,
            top_k=self.top_k,
            top_p=self.top_p,
            max_tokens=self.max_tokens,
            grammar=grammar,
        )
        return str(res["choices"][0]["message"]["content"])  # type: ignore



from pydantic import BaseModel, Field
import nodetool.metadata.types
from nodetool.metadata.types import *
from nodetool.dsl.graph import GraphNode


class GroupInput(GraphNode):
    items: list[Any] | GraphNode | tuple[GraphNode, str] = Field(default=[], description=None)
    name: str | GraphNode | tuple[GraphNode, str] = Field(default='', description=None)
    @classmethod
    def get_node_type(cls): return "nodetool.input.GroupInput"



class GroupNode(GraphNode):
    @classmethod
    def get_node_type(cls): return "nodetool.workflows.base_node.Group"



class GroupOutput(GraphNode):
    input: Any | GraphNode | tuple[GraphNode, str] = Field(default=None, description=None)
    name: str | GraphNode | tuple[GraphNode, str] = Field(default='', description=None)
    @classmethod
    def get_node_type(cls): return "nodetool.output.GroupOutput"



class Loop(GraphNode):
    @classmethod
    def get_node_type(cls): return "nodetool.group.Loop"





# ------------------------------------------------------------------------------
# Reference: https://github.com/facebookresearch/detectron2/blob/main/detectron2/evaluation/evaluator.py
# Modified by Jitesh Jain (https://github.com/praeclarumjj3)
# ------------------------------------------------------------------------------

import datetime
import logging
import time
from collections import OrderedDict, abc
from contextlib import ExitStack, contextmanager
from typing import List, Union
import torch
from torch import nn

from custom_detectron2.utils.comm import get_world_size, is_main_process
from custom_detectron2.utils.logger import log_every_n_seconds


class DatasetEvaluator:
    """
    Base class for a dataset evaluator.

    The function :func:`inference_on_dataset` runs the model over
    all samples in the dataset, and have a DatasetEvaluator to process the inputs/outputs.

    This class will accumulate information of the inputs/outputs (by :meth:`process`),
    and produce evaluation results in the end (by :meth:`evaluate`).
    """

    def reset(self):
        """
        Preparation for a new round of evaluation.
        Should be called before starting a round of evaluation.
        """
        pass

    def process(self, inputs, outputs):
        """
        Process the pair of inputs and outputs.
        If they contain batches, the pairs can be consumed one-by-one using `zip`:

        .. code-block:: python

            for input_, output in zip(inputs, outputs):
                # do evaluation on single input/output pair
                ...

        Args:
            inputs (list): the inputs that's used to call the model.
            outputs (list): the return value of `model(inputs)`
        """
        pass

    def evaluate(self):
        """
        Evaluate/summarize the performance, after processing all input/output pairs.

        Returns:
            dict:
                A new evaluator class can return a dict of arbitrary format
                as long as the user can process the results.
                In our train_net.py, we expect the following format:

                * key: the name of the task (e.g., bbox)
                * value: a dict of {metric name: score}, e.g.: {"AP50": 80}
        """
        pass


class DatasetEvaluators(DatasetEvaluator):
    """
    Wrapper class to combine multiple :class:`DatasetEvaluator` instances.

    This class dispatches every evaluation call to
    all of its :class:`DatasetEvaluator`.
    """

    def __init__(self, evaluators):
        """
        Args:
            evaluators (list): the evaluators to combine.
        """
        super().__init__()
        self._evaluators = evaluators

    def reset(self):
        for evaluator in self._evaluators:
            evaluator.reset()

    def process(self, inputs, outputs):
        for evaluator in self._evaluators:
            evaluator.process(inputs, outputs)

    def evaluate(self):
        results = OrderedDict()
        for evaluator in self._evaluators:
            result = evaluator.evaluate()
            if is_main_process() and result is not None:
                for k, v in result.items():
                    assert (
                        k not in results
                    ), "Different evaluators produce results with the same key {}".format(k)
                    results[k] = v
        return results


def inference_on_dataset(
    model, data_loader, evaluator: Union[DatasetEvaluator, List[DatasetEvaluator], None]
):
    """
    Run model on the data_loader and evaluate the metrics with evaluator.
    Also benchmark the inference speed of `model.__call__` accurately.
    The model will be used in eval mode.

    Args:
        model (callable): a callable which takes an object from
            `data_loader` and returns some outputs.

            If it's an nn.Module, it will be temporarily set to `eval` mode.
            If you wish to evaluate a model in `training` mode instead, you can
            wrap the given model and override its behavior of `.eval()` and `.train()`.
        data_loader: an iterable object with a length.
            The elements it generates will be the inputs to the model.
        evaluator: the evaluator(s) to run. Use `None` if you only want to benchmark,
            but don't want to do any evaluation.

    Returns:
        The return value of `evaluator.evaluate()`
    """
    num_devices = get_world_size()
    logger = logging.getLogger(__name__)
    logger.info("Start inference on {} batches".format(len(data_loader)))

    total = len(data_loader)  # inference data loader must have a fixed length
    if evaluator is None:
        # create a no-op evaluator
        evaluator = DatasetEvaluators([])
    if isinstance(evaluator, abc.MutableSequence):
        evaluator = DatasetEvaluators(evaluator)
    evaluator.reset()

    num_warmup = min(5, total - 1)
    start_time = time.perf_counter()
    total_data_time = 0
    total_compute_time = 0
    total_eval_time = 0
    with ExitStack() as stack:
        if isinstance(model, nn.Module):
            stack.enter_context(inference_context(model))
        stack.enter_context(torch.no_grad())

        start_data_time = time.perf_counter()
        for idx, inputs in enumerate(data_loader):
            total_data_time += time.perf_counter() - start_data_time
            if idx == num_warmup:
                start_time = time.perf_counter()
                total_data_time = 0
                total_compute_time = 0
                total_eval_time = 0

            start_compute_time = time.perf_counter()
            outputs = model(inputs)
            if torch.cuda.is_available():
                torch.cuda.synchronize()
            total_compute_time += time.perf_counter() - start_compute_time

            start_eval_time = time.perf_counter()
            evaluator.process(inputs, outputs)
            total_eval_time += time.perf_counter() - start_eval_time

            iters_after_start = idx + 1 - num_warmup * int(idx >= num_warmup)
            data_seconds_per_iter = total_data_time / iters_after_start
            compute_seconds_per_iter = total_compute_time / iters_after_start
            eval_seconds_per_iter = total_eval_time / iters_after_start
            total_seconds_per_iter = (time.perf_counter() - start_time) / iters_after_start
            if idx >= num_warmup * 2 or compute_seconds_per_iter > 5:
                eta = datetime.timedelta(seconds=int(total_seconds_per_iter * (total - idx - 1)))
                log_every_n_seconds(
                    logging.INFO,
                    (
                        f"Inference done {idx + 1}/{total}. "
                        f"Dataloading: {data_seconds_per_iter:.4f} s/iter. "
                        f"Inference: {compute_seconds_per_iter:.4f} s/iter. "
                        f"Eval: {eval_seconds_per_iter:.4f} s/iter. "
                        f"Total: {total_seconds_per_iter:.4f} s/iter. "
                        f"ETA={eta}"
                    ),
                    n=5,
                )
            start_data_time = time.perf_counter()

    # Measure the time only for this worker (before the synchronization barrier)
    total_time = time.perf_counter() - start_time
    total_time_str = str(datetime.timedelta(seconds=total_time))
    # NOTE this format is parsed by grep
    logger.info(
        "Total inference time: {} ({:.6f} s / iter per device, on {} devices)".format(
            total_time_str, total_time / (total - num_warmup), num_devices
        )
    )
    total_compute_time_str = str(datetime.timedelta(seconds=int(total_compute_time)))
    logger.info(
        "Total inference pure compute time: {} ({:.6f} s / iter per device, on {} devices)".format(
            total_compute_time_str, total_compute_time / (total - num_warmup), num_devices
        )
    )

    results = evaluator.evaluate()
    # An evaluator may return None when not in main process.
    # Replace it by an empty dict instead to make it easier for downstream code to handle
    if results is None:
        results = {}
    return results


@contextmanager
def inference_context(model):
    """
    A context where the model is temporarily changed to eval mode,
    and restored to previous mode afterwards.

    Args:
        model: a torch Module
    """
    training_mode = model.training
    model.eval()
    yield
    model.train(training_mode)



from enum import Enum
from pydantic import Field
from nodetool.metadata.types import Latent
from nodetool.common.comfy_node import ComfyNode


class Rotation(str, Enum):
    NONE = "none"
    _90_DEGREES = "90 degrees"
    _180_DEGREES = "180 degrees"
    _270_DEGREES = "270 degrees"


class LatentRotate(ComfyNode):
    samples: Latent = Field(
        default=Latent(), description="The latent samples to rotate."
    )
    rotation: Rotation = Field(
        default=Rotation.NONE, description="The degree of rotation to apply."
    )

    @classmethod
    def return_type(cls):
        return {"latent": Latent}


class FlipMethod(str, Enum):
    HORIZONTAL = "y-axis: horizontally"
    VERTICAL = "x-axis: vertically"


class LatentFlip(ComfyNode):
    samples: Latent = Field(default=Latent(), description="The latent samples to flip.")
    flip_method: FlipMethod = Field(
        default=FlipMethod.HORIZONTAL, description="The method to use for flipping."
    )

    @classmethod
    def return_type(cls):
        return {"latent": Latent}


class LatentCrop(ComfyNode):
    samples: Latent = Field(default=Latent(), description="The latent samples to crop.")
    width: int = Field(default=512, description="The width of the crop.")
    height: int = Field(default=512, description="The height of the crop.")
    x: int = Field(
        default=0,
        description="The x-coordinate for the top-left corner of the crop area.",
    )
    y: int = Field(
        default=0,
        description="The y-coordinate for the top-left corner of the crop area.",
    )

    @classmethod
    def return_type(cls):
        return {"latent": Latent}



from pydantic import BaseModel
from nodetool.metadata.types import Message, Task
from nodetool.models.message import MessageContent


class MessageCreateRequest(BaseModel):
    thread_id: str | None = None
    user_id: str | None = None
    tool_call_id: str | None = None
    role: str = ""
    name: str = ""
    content: str | list[MessageContent] | None = None
    tool_calls: list[dict] | None = None
    created_at: str | None = None


class MessageList(BaseModel):
    next: str | None
    messages: list[Message]


class TaskList(BaseModel):
    next: str | None
    tasks: list[Task]


class TaskCreateRequest(BaseModel):
    task_type: str
    thread_id: str
    name: str
    instructions: str
    dependencies: list[str] = []
    required_capabilities: list[str] = []


class TaskUpdateRequest(BaseModel):
    status: str | None = None
    error: str | None = None
    result: str | None = None
    cost: float | None = None
    started_at: str | None = None
    finished_at: str | None = None



from ..utils import common_annotator_call, create_node_input_types
import comfy.model_management as model_management
import os, sys
import subprocess, threading

#Ref: https://github.com/ltdrdata/ComfyUI-Manager/blob/284e90dc8296a2e1e4f14b4b2d10fba2f52f0e53/__init__.py#L14
def handle_stream(stream, prefix):
    for line in stream:
        print(prefix, line, end="")


def run_script(cmd, cwd='.'):
    process = subprocess.Popen(cmd, cwd=cwd, stdout=subprocess.PIPE, stderr=subprocess.PIPE, text=True, bufsize=1)

    stdout_thread = threading.Thread(target=handle_stream, args=(process.stdout, ""))
    stderr_thread = threading.Thread(target=handle_stream, args=(process.stderr, "[!]"))

    stdout_thread.start()
    stderr_thread.start()

    stdout_thread.join()
    stderr_thread.join()

    return process.wait()

def install_deps():
    try:
        import mediapipe
    except ImportError:
        run_script([sys.executable, '-s', '-m', 'pip', 'install', 'mediapipe'])
        run_script([sys.executable, '-s', '-m', 'pip', 'install', '--upgrade', 'protobuf'])

class Media_Pipe_Face_Mesh_Preprocessor:
    @classmethod
    def INPUT_TYPES(s):
        return create_node_input_types(
            max_faces=("INT", {"default": 10, "min": 1, "max": 50, "step": 1}), #Which image has more than 50 detectable faces?
            min_confidence=("FLOAT", {"default": 0.5, "min": 0.01, "max": 1.0, "step": 0.01})
        )
        
    RETURN_TYPES = ("IMAGE",)
    FUNCTION = "detect"

    CATEGORY = "ControlNet Preprocessors/Faces and Poses Estimators"

    def detect(self, image, max_faces, min_confidence, resolution=512):
        #Ref: https://github.com/Fannovel16/comfy_controlnet_preprocessors/issues/70#issuecomment-1677967369
        install_deps()
        from controlnet_aux.mediapipe_face import MediapipeFaceDetector
        return (common_annotator_call(MediapipeFaceDetector(), image, max_faces=max_faces, min_confidence=min_confidence, resolution=resolution), )

NODE_CLASS_MAPPINGS = {
    "MediaPipe-FaceMeshPreprocessor": Media_Pipe_Face_Mesh_Preprocessor
}

NODE_DISPLAY_NAME_MAPPINGS = {
    "MediaPipe-FaceMeshPreprocessor": "MediaPipe Face Mesh"
}


import os
from .inference import init_segmentor, inference_segmentor, show_result_pyplot
import warnings
import cv2
import numpy as np
from PIL import Image
from controlnet_aux.util import HWC3, common_input_validate, resize_image_with_pad, custom_hf_download, HF_MODEL_NAME
import torch

from custom_mmpkg.custom_mmseg.core.evaluation import get_palette

config_file = os.path.join(os.path.dirname(os.path.realpath(__file__)),  "upernet_global_small.py")



class UniformerSegmentor:
    def __init__(self, netNetwork):
        self.model = netNetwork
    
    @classmethod
    def from_pretrained(cls, pretrained_model_or_path=HF_MODEL_NAME, filename="upernet_global_small.pth"):
        model_path = custom_hf_download(pretrained_model_or_path, filename)

        netNetwork = init_segmentor(config_file, model_path, device="cpu")
        netNetwork.load_state_dict({k.replace('module.', ''): v for k, v in torch.load(model_path)['state_dict'].items()})
        netNetwork.eval()

        return cls(netNetwork)

    def to(self, device):
        self.model.to(device)
        return self

    def _inference(self, img):
        if next(self.model.parameters()).device.type == 'mps':
            # adaptive_avg_pool2d can fail on MPS, workaround with CPU
            import torch.nn.functional
            
            orig_adaptive_avg_pool2d = torch.nn.functional.adaptive_avg_pool2d
            def cpu_if_exception(input, *args, **kwargs):
                try:
                    return orig_adaptive_avg_pool2d(input, *args, **kwargs)
                except:
                    return orig_adaptive_avg_pool2d(input.cpu(), *args, **kwargs).to(input.device)
            
            try:
                torch.nn.functional.adaptive_avg_pool2d = cpu_if_exception
                result = inference_segmentor(self.model, img)
            finally:
                torch.nn.functional.adaptive_avg_pool2d = orig_adaptive_avg_pool2d
        else:
            result = inference_segmentor(self.model, img)

        res_img = show_result_pyplot(self.model, img, result, get_palette('ade'), opacity=1)
        return res_img

    def __call__(self, input_image=None, detect_resolution=512, output_type=None, upscale_method="INTER_CUBIC", **kwargs):
        input_image, output_type = common_input_validate(input_image, output_type, **kwargs)
        input_image, remove_pad = resize_image_with_pad(input_image, detect_resolution, upscale_method)

        detected_map = self._inference(input_image)
        detected_map = remove_pad(HWC3(detected_map))
        
        if output_type == "pil":
            detected_map = Image.fromarray(detected_map)
            
        return detected_map




from pydantic import BaseModel, Field
import nodetool.metadata.types
from nodetool.metadata.types import *
from nodetool.dsl.graph import GraphNode


class SaveAnimatedPNG(GraphNode):
    images: list[ImageTensor] | GraphNode | tuple[GraphNode, str] = Field(default=PydanticUndefined, description='list of images to save as animated PNG.')
    filename_prefix: str | GraphNode | tuple[GraphNode, str] = Field(default='ComfyUI', description='Prefix for the filename.')
    fps: float | GraphNode | tuple[GraphNode, str] = Field(default=6.0, description='Frames per second.')
    compress_level: int | GraphNode | tuple[GraphNode, str] = Field(default=4, description='PNG compression level.')
    hidden_fields: dict[str, Any] | GraphNode | tuple[GraphNode, str] = Field(default=PydanticUndefined, description='Hidden fields like prompt and extra PNG information.')
    output_node: bool | GraphNode | tuple[GraphNode, str] = Field(default=True, description=None)
    @classmethod
    def get_node_type(cls): return "comfy.image.animation.SaveAnimatedPNG"


from enum import Enum

class SaveAnimatedWEBP(GraphNode):
    images: list[ImageTensor] | GraphNode | tuple[GraphNode, str] = Field(default=PydanticUndefined, description='list of images to save as animated WEBP.')
    filename_prefix: str | GraphNode | tuple[GraphNode, str] = Field(default='ComfyUI', description='Prefix for the filename.')
    fps: float | GraphNode | tuple[GraphNode, str] = Field(default=6.0, description='Frames per second.')
    lossless: bool | GraphNode | tuple[GraphNode, str] = Field(default=True, description='Whether to use lossless compression.')
    quality: int | GraphNode | tuple[GraphNode, str] = Field(default=80, description='Quality of the WEBP.')
    method: Enum | GraphNode | tuple[GraphNode, str] = Field(default='default', description='Compression method to use.')
    hidden_fields: dict[str, Any] | GraphNode | tuple[GraphNode, str] = Field(default=PydanticUndefined, description='Hidden fields like prompt and extra PNG information.')
    output_node: bool | GraphNode | tuple[GraphNode, str] = Field(default=True, description=None)
    @classmethod
    def get_node_type(cls): return "comfy.image.animation.SaveAnimatedWEBP"





from pydantic import BaseModel, Field
import nodetool.metadata.types
from nodetool.metadata.types import *
from nodetool.dsl.graph import GraphNode


class Agent(GraphNode):
    model: FunctionModel | GraphNode | tuple[GraphNode, str] = Field(default=FunctionModel(type='function_model', name='', repo_id='', filename='', local_path=None), description='The language model to use.')
    goal: str | GraphNode | tuple[GraphNode, str] = Field(default='', description='The user prompt')
    n_gpu_layers: int | GraphNode | tuple[GraphNode, str] = Field(default=0, description='Number of layers on the GPU')
    @classmethod
    def get_node_type(cls): return "nodetool.agents.Agent"



class ImageTask(GraphNode):
    model: FunctionModel | GraphNode | tuple[GraphNode, str] = Field(default=FunctionModel(type='function_model', name='gpt-4-1106-preview', repo_id='', filename='', local_path=None), description='The language model to use.')
    thread_id: str | GraphNode | tuple[GraphNode, str] = Field(default='', description='The thread ID of the conversation.')
    tasks: list[Task] | GraphNode | tuple[GraphNode, str] = Field(default=[], description='The tasks to be executed by this agent.')
    nodes: list[NodeRef] | GraphNode | tuple[GraphNode, str] = Field(default=[], description='The nodes to use as tools.')
    @classmethod
    def get_node_type(cls): return "nodetool.agents.ImageTask"



class TaskNode(GraphNode):
    model: FunctionModel | GraphNode | tuple[GraphNode, str] = Field(default=FunctionModel(type='function_model', name='gpt-4-1106-preview', repo_id='', filename='', local_path=None), description='The language model to use.')
    thread_id: str | GraphNode | tuple[GraphNode, str] = Field(default='', description='The thread ID of the conversation.')
    tasks: list[Task] | GraphNode | tuple[GraphNode, str] = Field(default=[], description='The tasks to be executed by this agent.')
    nodes: list[NodeRef] | GraphNode | tuple[GraphNode, str] = Field(default=[], description='The nodes to use as tools.')
    @classmethod
    def get_node_type(cls): return "nodetool.agents.Task"



class TextTasks(GraphNode):
    model: FunctionModel | GraphNode | tuple[GraphNode, str] = Field(default=FunctionModel(type='function_model', name='gpt-4-1106-preview', repo_id='', filename='', local_path=None), description='The language model to use.')
    thread_id: str | GraphNode | tuple[GraphNode, str] = Field(default='', description='The thread ID of the conversation.')
    tasks: list[Task] | GraphNode | tuple[GraphNode, str] = Field(default=[], description='The tasks to be executed by this agent.')
    nodes: list[NodeRef] | GraphNode | tuple[GraphNode, str] = Field(default=[], description='The nodes to use as tools.')
    @classmethod
    def get_node_type(cls): return "nodetool.agents.TextTasks"





from pydantic import BaseModel, Field
import nodetool.metadata.types
from nodetool.metadata.types import *
from nodetool.dsl.graph import GraphNode

from nodetool.nodes.openai.text import EmbeddingModel

class Embedding(GraphNode):
    input: str | nodetool.metadata.types.TextRef | GraphNode | tuple[GraphNode, str] = Field(default='', description=None)
    model: EmbeddingModel | GraphNode | tuple[GraphNode, str] = Field(default=EmbeddingModel('text-embedding-3-small'), description=None)
    chunk_size: int | GraphNode | tuple[GraphNode, str] = Field(default=4096, description=None)
    @classmethod
    def get_node_type(cls): return "openai.text.Embedding"


from nodetool.metadata.types import GPTModel
from nodetool.nodes.openai.text import ResponseFormat

class GPT(GraphNode):
    model: GPTModel | GraphNode | tuple[GraphNode, str] = Field(default=GPTModel('gpt-3.5-turbo-0125'), description=None)
    system: str | GraphNode | tuple[GraphNode, str] = Field(default='You are a friendly assistant.', description=None)
    prompt: str | GraphNode | tuple[GraphNode, str] = Field(default='', description=None)
    image: ImageRef | GraphNode | tuple[GraphNode, str] = Field(default=ImageRef(type='image', uri='', asset_id=None, temp_id=None), description=None)
    presence_penalty: float | GraphNode | tuple[GraphNode, str] = Field(default=0.0, description=None)
    frequency_penalty: float | GraphNode | tuple[GraphNode, str] = Field(default=0.0, description=None)
    temperature: float | GraphNode | tuple[GraphNode, str] = Field(default=1.0, description=None)
    max_tokens: int | GraphNode | tuple[GraphNode, str] = Field(default=100, description=None)
    top_p: float | GraphNode | tuple[GraphNode, str] = Field(default=1.0, description=None)
    response_format: ResponseFormat | GraphNode | tuple[GraphNode, str] = Field(default=ResponseFormat('text'), description=None)
    @classmethod
    def get_node_type(cls): return "openai.text.GPT"





import base64
from io import BytesIO
from typing import Any
import pydub

from openai.types.completion_usage import CompletionUsage
from openai.types.create_embedding_response import Usage
from openai.types.chat import ChatCompletion, ChatCompletionMessageParam

from nodetool.common.environment import Environment
from nodetool.models.prediction import Prediction


def calculate_cost(
    model: str,
    input_tokens: int,
    output_tokens: int | None = None,
) -> float:
    # Pricing information for different models
    pricing: dict[str, Any] = {
        "gpt-4o": {"input": 5.00, "output": 15.00},
        "gpt-3.5-turbo-0125": {"input": 0.50, "output": 1.50},
        "gpt-3.5-turbo-instruct": {"input": 1.50, "output": 2.00},
        "text-embedding-3-small": {"usage": 0.02},
        "text-embedding-3-large": {"usage": 0.13},
        "ada v2": {"usage": 0.10},
        "whisper-1": {"usage": 0.0001},
        "tts-1": {"usage": 15.00},
        "tts-1-hd": {"usage": 30.00},
        "dalle-3 standard 1024x1024": {"price": 0.040},
        "dalle-3 standard 1024x1792, 1792x1024": {"price": 0.080},
        "dalle-3 hd 1024x1024": {"price": 0.080},
        "dalle-3 hd 1024x1792, 1792x1024": {"price": 0.120},
    }

    if model.lower() in pricing:
        model_pricing = pricing[model.lower()]

        if "input" in model_pricing and "output" in model_pricing:
            input_price = model_pricing["input"] / 1_000_000 * input_tokens
            output_price = model_pricing["output"] / 1_000_000 * output_tokens
            total_price = input_price + output_price
        elif "usage" in model_pricing:
            total_price = model_pricing["usage"] / 1_000_000 * input_tokens
        elif "price" in model_pricing:
            total_price = model_pricing["price"] * 1
        else:
            raise ValueError(f"Invalid pricing structure for model: {model}")

        return total_price
    else:
        raise ValueError(f"Unsupported model")


def calculate_cost_for_completion_usage(model: str, usage: CompletionUsage):
    return calculate_cost(model, usage.prompt_tokens, usage.completion_tokens)


def calculate_cost_for_embedding_usage(model: str, usage: Usage):
    return calculate_cost(model, usage.prompt_tokens)


async def create_embedding(prediction: Prediction, input: str) -> Any:
    client = Environment.get_openai_client()
    res = await client.embeddings.create(input=input, model=prediction.model)
    cost = calculate_cost_for_embedding_usage(prediction.model, res.usage)
    prediction.cost = cost
    prediction.save()
    return res.model_dump()


async def create_speech(prediction: Prediction, params: dict) -> Any:
    client = Environment.get_openai_client()
    res = await client.audio.speech.create(
        model=prediction.model,
        input=params["input"],
        voice=params["voice"],
        speed=params["speed"],
        response_format="mp3",
    )
    prediction.cost = calculate_cost(prediction.model, len(params["input"]), 0)
    prediction.save()
    return res.content


async def create_chat_completion(
    prediction: Prediction,
    params: dict[str, Any],
) -> Any:
    client = Environment.get_openai_client()

    res: ChatCompletion = await client.chat.completions.create(
        model=prediction.model,
        messages=params["messages"],
        max_tokens=params["max_tokens"],
        temperature=params["temperature"],
        top_p=params["top_p"],
        presence_penalty=params["presence_penalty"],
        frequency_penalty=params["frequency_penalty"],
        response_format={"type": params["response_format"]},
    )
    assert res.usage is not None
    prediction.cost = calculate_cost_for_completion_usage(prediction.model, res.usage)
    prediction.save()
    return res.model_dump()


async def create_whisper(
    prediction: Prediction,
    params: dict,
) -> Any:
    client = Environment.get_openai_client()
    file = base64.b64decode(params["file"])
    audio_segment: pydub.AudioSegment = pydub.AudioSegment.from_file(BytesIO(file))

    if params.get("translate", False):
        res = await client.audio.translations.create(
            model=prediction.model,
            file=("file.mp3", file, "audio/mp3"),
            temperature=params["temperature"],
        )
    else:
        res = await client.audio.transcriptions.create(
            model=prediction.model,
            file=("file.mp3", file, "audio/mp3"),
            temperature=params["temperature"] if "temperature" in params else 0.0,
        )

    prediction.cost = calculate_cost(
        prediction.model, int(audio_segment.duration_seconds)
    )
    prediction.save()

    return res.model_dump()


async def create_image(prediction: Prediction, params: dict) -> Any:
    client = Environment.get_openai_client()
    image = await client.images.generate(
        model=prediction.model,
        prompt=params["prompt"],
        size=params["size"],
        quality=params["quality"],
        style=params["style"],
        response_format="b64_json",
    )
    cost_model = f"dalle-3 {params['quality']} {params['size']}"
    prediction.cost = calculate_cost(cost_model, len(params["prompt"]), 0)
    prediction.save()

    return image.model_dump()


async def run_openai(prediction: Prediction, params: dict) -> Any:
    client = Environment.get_openai_client()
    model = prediction.model

    if model.startswith("text-embedding-"):
        return await create_embedding(prediction, params["input"])

    elif model.startswith("gpt-"):
        return await create_chat_completion(prediction, params)

    elif model.startswith("tts-"):
        return await create_speech(prediction, params)

    elif model.startswith("whisper-"):
        return await create_whisper(prediction, params)

    elif model.startswith("dall-e-"):
        return await create_image(prediction, params)

    else:
        raise ValueError(f"Unsupported model: {model}")



from pydantic import BaseModel, Field
import nodetool.metadata.types
from nodetool.metadata.types import *
from nodetool.dsl.graph import GraphNode


class HyperTile(GraphNode):
    model: UNet | GraphNode | tuple[GraphNode, str] = Field(default=UNet(type='comfy.unet'), description='The model to use for generating hyper-tiles.')
    tile_size: int | GraphNode | tuple[GraphNode, str] = Field(default=256, description='The size of the tile to generate.')
    swap_size: int | GraphNode | tuple[GraphNode, str] = Field(default=2, description='The swap size used during generation.')
    max_depth: int | GraphNode | tuple[GraphNode, str] = Field(default=0, description='The maximum depth for tiling.')
    scale_depth: bool | GraphNode | tuple[GraphNode, str] = Field(default=False, description='Whether to scale the depth progressively.')
    @classmethod
    def get_node_type(cls): return "comfy._for_testing.HyperTile"





import nodetool.nodes.comfy
import nodetool.nodes.comfy.advanced
import nodetool.nodes.comfy.advanced.conditioning
import nodetool.nodes.comfy.advanced.loaders
import nodetool.nodes.comfy.conditioning
import nodetool.nodes.comfy.controlnet
import nodetool.nodes.comfy.controlnet.faces_and_poses
import nodetool.nodes.comfy.controlnet.semantic_segmentation
import nodetool.nodes.comfy.controlnet.normal_and_depth
import nodetool.nodes.comfy.controlnet.others
import nodetool.nodes.comfy.controlnet.line_extractors
import nodetool.nodes.comfy.controlnet.t2i
import nodetool.nodes.comfy.generate
import nodetool.nodes.comfy.image
import nodetool.nodes.comfy.image.animation
import nodetool.nodes.comfy.image.batch
import nodetool.nodes.comfy.image.preprocessors
import nodetool.nodes.comfy.image.transform
import nodetool.nodes.comfy.image.upscaling
import nodetool.nodes.comfy.ipadapter
import nodetool.nodes.comfy.latent
import nodetool.nodes.comfy.latent.advanced
import nodetool.nodes.comfy.latent.batch
import nodetool.nodes.comfy.latent.inpaint
import nodetool.nodes.comfy.latent.transform
import nodetool.nodes.comfy.loaders
import nodetool.nodes.comfy.mask
import nodetool.nodes.comfy.mask.compositing
import nodetool.nodes.comfy.sampling
import nodetool.nodes.comfy.sampling.samplers
import nodetool.nodes.comfy.sampling.schedulers
import nodetool.nodes.comfy.sampling.sigmas



from pydantic import BaseModel, Field
import nodetool.metadata.types
from nodetool.metadata.types import *
from nodetool.dsl.graph import GraphNode

from nodetool.nodes.comfy.sampling import SchedulerEnum

class BasicScheduler(GraphNode):
    model: UNet | GraphNode | tuple[GraphNode, str] = Field(default=UNet(type='comfy.unet'), description='The model to use.')
    scheduler: SchedulerEnum | GraphNode | tuple[GraphNode, str] = Field(default=SchedulerEnum('normal'), description='The scheduler name.')
    steps: int | GraphNode | tuple[GraphNode, str] = Field(default=20, description='The number of steps.')
    @classmethod
    def get_node_type(cls): return "comfy.sampling.schedulers.BasicScheduler"



class ExponentialScheduler(GraphNode):
    steps: int | GraphNode | tuple[GraphNode, str] = Field(default=20, description='The number of steps.')
    sigma_max: float | GraphNode | tuple[GraphNode, str] = Field(default=14.614642, description='The maximum sigma value.')
    sigma_min: float | GraphNode | tuple[GraphNode, str] = Field(default=0.0291675, description='The minimum sigma value.')
    @classmethod
    def get_node_type(cls): return "comfy.sampling.schedulers.ExponentialScheduler"



class KarrasScheduler(GraphNode):
    steps: int | GraphNode | tuple[GraphNode, str] = Field(default=20, description='The number of steps.')
    sigma_max: float | GraphNode | tuple[GraphNode, str] = Field(default=14.614642, description='The maximum sigma value.')
    sigma_min: float | GraphNode | tuple[GraphNode, str] = Field(default=0.0291675, description='The minimum sigma value.')
    rho: float | GraphNode | tuple[GraphNode, str] = Field(default=7.0, description='The rho value.')
    @classmethod
    def get_node_type(cls): return "comfy.sampling.schedulers.KarrasScheduler"



class PolyexponentialScheduler(GraphNode):
    steps: int | GraphNode | tuple[GraphNode, str] = Field(default=20, description='The number of steps to compute the sigmas.')
    sigma_max: float | GraphNode | tuple[GraphNode, str] = Field(default=14.614642, description='The maximum sigma value.')
    sigma_min: float | GraphNode | tuple[GraphNode, str] = Field(default=0.0291675, description='The minimum sigma value.')
    rho: float | GraphNode | tuple[GraphNode, str] = Field(default=1.0, description='The rho parameter for the scheduler.')
    @classmethod
    def get_node_type(cls): return "comfy.sampling.schedulers.PolyexponentialScheduler"



class SDTurboScheduler(GraphNode):
    model: UNet | GraphNode | tuple[GraphNode, str] = Field(default=UNet(type='comfy.unet'), description='The model for which to use the scheduler.')
    steps: int | GraphNode | tuple[GraphNode, str] = Field(default=1, description='The number of steps for the scheduler.')
    denoise: float | GraphNode | tuple[GraphNode, str] = Field(default=1.0, description='The denoising factor to apply in the scheduler.')
    @classmethod
    def get_node_type(cls): return "comfy.sampling.schedulers.SDTurboScheduler"



class VPScheduler(GraphNode):
    steps: int | GraphNode | tuple[GraphNode, str] = Field(default=20, description='The number of steps to compute the sigmas.')
    beta_d: float | GraphNode | tuple[GraphNode, str] = Field(default=19.9, description='beta_d parameter for the VP scheduler.')
    beta_min: float | GraphNode | tuple[GraphNode, str] = Field(default=0.1, description='beta_min parameter for the VP scheduler.')
    eps_s: float | GraphNode | tuple[GraphNode, str] = Field(default=0.001, description='eps_s parameter for the VP scheduler.')
    @classmethod
    def get_node_type(cls): return "comfy.sampling.schedulers.VPScheduler"





from pydantic import Field
from nodetool.metadata.types import Sigmas, UNet
from nodetool.common.comfy_node import ComfyNode
from nodetool.nodes.comfy.sampling import SchedulerEnum


class BasicScheduler(ComfyNode):
    model: UNet = Field(default=UNet(), description="The model to use.")
    scheduler: SchedulerEnum = Field(
        default=SchedulerEnum.NORMAL, description="The scheduler name."
    )
    steps: int = Field(default=20, description="The number of steps.")

    @classmethod
    def return_type(cls):
        return {"sigmas": Sigmas}


class KarrasScheduler(ComfyNode):
    steps: int = Field(default=20, description="The number of steps.")
    sigma_max: float = Field(default=14.614642, description="The maximum sigma value.")
    sigma_min: float = Field(default=0.0291675, description="The minimum sigma value.")
    rho: float = Field(default=7.0, description="The rho value.")

    @classmethod
    def return_type(cls):
        return {"sigmas": Sigmas}


class ExponentialScheduler(ComfyNode):
    steps: int = Field(default=20, description="The number of steps.")
    sigma_max: float = Field(default=14.614642, description="The maximum sigma value.")
    sigma_min: float = Field(default=0.0291675, description="The minimum sigma value.")

    @classmethod
    def return_type(cls):
        return {"sigmas": Sigmas}


class PolyexponentialScheduler(ComfyNode):
    steps: int = Field(
        default=20, description="The number of steps to compute the sigmas."
    )
    sigma_max: float = Field(default=14.614642, description="The maximum sigma value.")
    sigma_min: float = Field(default=0.0291675, description="The minimum sigma value.")
    rho: float = Field(default=1.0, description="The rho parameter for the scheduler.")

    @classmethod
    def return_type(cls):
        return {"sigmas": Sigmas}


class SDTurboScheduler(ComfyNode):
    model: UNet = Field(
        default=UNet(), description="The model for which to use the scheduler."
    )
    steps: int = Field(default=1, description="The number of steps for the scheduler.")
    denoise: float = Field(
        default=1.0, description="The denoising factor to apply in the scheduler."
    )

    @classmethod
    def return_type(cls):
        return {"sigmas": Sigmas}


class VPScheduler(ComfyNode):
    steps: int = Field(
        default=20, description="The number of steps to compute the sigmas."
    )
    beta_d: float = Field(
        default=19.9, description="beta_d parameter for the VP scheduler."
    )
    beta_min: float = Field(
        default=0.1, description="beta_min parameter for the VP scheduler."
    )
    eps_s: float = Field(
        default=0.001, description="eps_s parameter for the VP scheduler."
    )

    @classmethod
    def return_type(cls):
        return {"sigmas": Sigmas}



import torch


def batch_rotprojs(batches_rotmats):
    proj_rotmats = []
    for batch_idx, batch_rotmats in enumerate(batches_rotmats):
        proj_batch_rotmats = []
        for rot_idx, rotmat in enumerate(batch_rotmats):
            # GPU implementation of svd is VERY slow
            # ~ 2 10^-3 per hit vs 5 10^-5 on cpu
            U, S, V = rotmat.cpu().svd()
            rotmat = torch.matmul(U, V.transpose(0, 1))
            orth_det = rotmat.det()
            # Remove reflection
            if orth_det < 0:
                rotmat[:, 2] = -1 * rotmat[:, 2]

            rotmat = rotmat.cuda()
            proj_batch_rotmats.append(rotmat)
        proj_rotmats.append(torch.stack(proj_batch_rotmats))
    return torch.stack(proj_rotmats)



from ..utils import common_annotator_call, create_node_input_types
import comfy.model_management as model_management

class Lineart_Standard_Preprocessor:
    @classmethod
    def INPUT_TYPES(s):
        return create_node_input_types(
            guassian_sigma=("FLOAT", {"default": 6.0, "min": 0.0, "max": 100.0}),
            intensity_threshold=("INT", {"default": 8, "min": 0, "max": 16})
        )

    RETURN_TYPES = ("IMAGE",)
    FUNCTION = "execute"

    CATEGORY = "ControlNet Preprocessors/Line Extractors"

    def execute(self, image, guassian_sigma, intensity_threshold, resolution=512, **kwargs):
        from controlnet_aux.lineart_standard import LineartStandardDetector
        return (common_annotator_call(LineartStandardDetector(), image, guassian_sigma=guassian_sigma, intensity_threshold=intensity_threshold, resolution=resolution), )

NODE_CLASS_MAPPINGS = {
    "LineartStandardPreprocessor": Lineart_Standard_Preprocessor
}
NODE_DISPLAY_NAME_MAPPINGS = {
    "LineartStandardPreprocessor": "Standard Lineart"
}


from pydantic import BaseModel, Field
import nodetool.metadata.types
from nodetool.metadata.types import *
from nodetool.dsl.graph import GraphNode

from nodetool.nodes.comfy.controlnet.faces_and_poses import Toggle
from nodetool.nodes.comfy.controlnet.faces_and_poses import Toggle
from nodetool.nodes.comfy.controlnet.faces_and_poses import Toggle
from nodetool.nodes.comfy.controlnet.faces_and_poses import BBoxDetectorModel
from nodetool.nodes.comfy.controlnet.faces_and_poses import PoseEstimatorModel


class DWPose_Preprocessor(GraphNode):
    detect_hand: Toggle | GraphNode | tuple[GraphNode, str] = Field(
        default=Toggle("enable"),
        description="Toggle to enable or disable hand detection.",
    )
    detect_body: Toggle | GraphNode | tuple[GraphNode, str] = Field(
        default=Toggle("enable"),
        description="Toggle to enable or disable body detection.",
    )
    detect_face: Toggle | GraphNode | tuple[GraphNode, str] = Field(
        default=Toggle("enable"),
        description="Toggle to enable or disable face detection.",
    )
    bbox_detector: BBoxDetectorModel | GraphNode | tuple[GraphNode, str] = Field(
        default=BBoxDetectorModel("yolox_l.torchscript.pt"),
        description="The bounding box detector model to use.",
    )
    pose_estimator: PoseEstimatorModel | GraphNode | tuple[GraphNode, str] = Field(
        default=PoseEstimatorModel("dw-ll_ucoco_384_bs5.torchscript.pt"),
        description="The pose estimator model to use.",
    )

    @classmethod
    def get_node_type(cls):
        return "comfy.controlnet.faces_and_poses.DWPose_Preprocessor"


from nodetool.common.comfy_node import DensePoseModel
from nodetool.nodes.comfy.controlnet.faces_and_poses import DensePoseCMap


class DensePosePreprocessor(GraphNode):
    image: ImageTensor | GraphNode | tuple[GraphNode, str] = Field(
        default=ImageTensor(type="comfy.image_tensor"),
        description="The image to preprocess.",
    )
    resolution: int | GraphNode | tuple[GraphNode, str] = Field(
        default=512, description="The width of the image to generate."
    )
    model: DensePoseModel | GraphNode | tuple[GraphNode, str] = Field(
        default=DensePoseModel("densepose_r50_fpn_dl.torchscript"),
        description="The model to use.",
    )
    cmap: DensePoseCMap | GraphNode | tuple[GraphNode, str] = Field(
        default=DensePoseCMap("Viridis (MagicAnimate)"),
        description="The color map to use.",
    )

    @classmethod
    def get_node_type(cls):
        return "comfy.controlnet.faces_and_poses.DensePosePreprocessor"


class MediaPipeFaceMeshPreprocessor(GraphNode):
    image: ImageTensor | GraphNode | tuple[GraphNode, str] = Field(
        default=ImageTensor(type="comfy.image_tensor"),
        description="The image to preprocess.",
    )
    resolution: int | GraphNode | tuple[GraphNode, str] = Field(
        default=512, description="The width of the image to generate."
    )
    max_faces: int | GraphNode | tuple[GraphNode, str] = Field(
        default=10, description="The maximum number of faces to detect."
    )
    min_confidence: float | GraphNode | tuple[GraphNode, str] = Field(
        default=0.5, description="The minimum confidence for face detection."
    )

    @classmethod
    def get_node_type(cls):
        return "comfy.controlnet.faces_and_poses.MediaPipeFaceMeshPreprocessor"


from nodetool.common.comfy_node import EnableDisable
from nodetool.common.comfy_node import EnableDisable
from nodetool.common.comfy_node import EnableDisable


class OpenposePreprocessor(GraphNode):
    image: ImageTensor | GraphNode | tuple[GraphNode, str] = Field(
        default=ImageTensor(type="comfy.image_tensor"),
        description="The image to preprocess.",
    )
    resolution: int | GraphNode | tuple[GraphNode, str] = Field(
        default=512, description="The width of the image to generate."
    )
    detect_hand: EnableDisable | GraphNode | tuple[GraphNode, str] = Field(
        default=EnableDisable("enable"), description="Whether to detect hands."
    )
    detect_body: EnableDisable | GraphNode | tuple[GraphNode, str] = Field(
        default=EnableDisable("enable"), description="Whether to detect bodies."
    )
    detect_face: EnableDisable | GraphNode | tuple[GraphNode, str] = Field(
        default=EnableDisable("enable"), description="Whether to detect faces."
    )

    @classmethod
    def get_node_type(cls):
        return "comfy.controlnet.faces_and_poses.OpenposePreprocessor"


class PreprocessImage(GraphNode):
    image: ImageTensor | GraphNode | tuple[GraphNode, str] = Field(
        default=ImageTensor(type="comfy.image_tensor"),
        description="The image to preprocess.",
    )
    resolution: int | GraphNode | tuple[GraphNode, str] = Field(
        default=512, description="The width of the image to generate."
    )

    @classmethod
    def get_node_type(cls):
        return "comfy.controlnet.PreprocessImage"



# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
# Copyright (c) @comfyanonymous
# Project Repository: https://github.com/comfyanonymous/ComfyUI

import os
from comfy.comfy_extras.chainner_models import model_loading
from comfy import model_management
import torch
import comfy.utils
import comfy.folder_paths as folder_paths


class UpscaleModelLoader:
    @classmethod
    def INPUT_TYPES(s):
        return {
            "required": {
                "model_name": (folder_paths.get_filename_list("upscale_models"),),
            }
        }

    RETURN_TYPES = ("UPSCALE_MODEL",)
    FUNCTION = "load_model"

    CATEGORY = "loaders"

    def load_model(self, model_name):
        model_path = folder_paths.get_full_path("upscale_models", model_name)
        sd = comfy.utils.load_torch_file(model_path, safe_load=True)
        if "module.layers.0.residual_group.blocks.0.norm1.weight" in sd:
            sd = comfy.utils.state_dict_prefix_replace(sd, {"module.": ""})
        out = model_loading.load_state_dict(sd).eval()
        return (out,)


class ImageUpscaleWithModel:
    @classmethod
    def INPUT_TYPES(s):
        return {
            "required": {
                "upscale_model": ("UPSCALE_MODEL",),
                "image": ("IMAGE",),
            }
        }

    RETURN_TYPES = ("IMAGE",)
    FUNCTION = "upscale"

    CATEGORY = "image/upscaling"

    def upscale(self, upscale_model, image):
        device = model_management.get_torch_device()
        upscale_model.to(device)
        in_img = image.movedim(-1, -3).to(device)
        free_memory = model_management.get_free_memory(device)

        tile = 512
        overlap = 32

        oom = True
        while oom:
            try:
                steps = in_img.shape[0] * comfy.utils.get_tiled_scale_steps(
                    in_img.shape[3],
                    in_img.shape[2],
                    tile_x=tile,
                    tile_y=tile,
                    overlap=overlap,
                )
                pbar = comfy.utils.ProgressBar(steps)
                s = comfy.utils.tiled_scale(
                    in_img,
                    lambda a: upscale_model(a),
                    tile_x=tile,
                    tile_y=tile,
                    overlap=overlap,
                    upscale_amount=upscale_model.scale,
                    pbar=pbar,
                )
                oom = False
            except model_management.OOM_EXCEPTION as e:
                tile //= 2
                if tile < 128:
                    raise e

        upscale_model.cpu()
        s = torch.clamp(s.movedim(-3, -1), min=0, max=1.0)
        return (s,)


NODE_CLASS_MAPPINGS = {
    "UpscaleModelLoader": UpscaleModelLoader,
    "ImageUpscaleWithModel": ImageUpscaleWithModel,
}



import base64
from enum import Enum
from nodetool.workflows.base_node import BaseNode
from pydantic import Field
from io import BytesIO
from pydub import AudioSegment
from nodetool.metadata.types import AudioRef
from nodetool.workflows.processing_context import ProcessingContext
from openai.types.audio.transcription import Transcription
from openai.types.audio.translation import Translation


class TextToSpeech(BaseNode):
    """
    Converts text into spoken voice using OpenAI TTS models.
    audio, tts, t2s, text-to-speech, voiceover, speak, voice, read
    Returns an audio file from the provided text.
    """

    class TtsModel(str, Enum):
        tts_1 = "tts-1"
        tts_1_hd = "tts-1-hd"

    class Voice(str, Enum):
        ALLOY = "alloy"
        ECHO = "echo"
        FABLE = "fable"
        ONYX = "onyx"
        NOVA = "nova"
        SHIMMER = "shimmer"

    model: TtsModel = Field(title="Model", default=TtsModel.tts_1)
    voice: Voice = Field(title="Voice", default=Voice.ALLOY)
    input: str = Field(title="Input", default="")
    speed: float = Field(title="Speed", default=1.0, ge=0.25, le=4.0)

    async def process(self, context: ProcessingContext) -> AudioRef:
        res = await context.run_prediction(
            node_id=self._id,
            provider="openai",
            model=self.model.value,
            params={
                "input": self.input,
                "voice": self.voice,
                "speed": self.speed,
            },
        )

        segment = AudioSegment.from_mp3(BytesIO(res.content))
        audio = await context.audio_from_segment(segment)  # type: ignore
        return audio


class Transcribe(BaseNode):
    """
    Converts spoken words in an audio file to written text.
    audio, stt, s2t, speech-to-text, transcription, audio-to-text, analysis
    Returns a text of the detected words from the input audio file.
    """

    audio: AudioRef = Field(
        default=AudioRef(), description="The audio file to transcribe."
    )
    temperature: float = Field(
        default=0.0, description="The temperature to use for the transcription."
    )

    async def process(self, context: ProcessingContext) -> str:
        audio_bytes = await context.asset_to_io(self.audio)

        response = await context.run_prediction(
            node_id=self._id,
            provider="openai",
            model="whisper-1",
            params={
                "file": base64.b64encode(audio_bytes.read()).decode(),
            },
        )

        res = Transcription(**response.json())

        return res.text


class Translate(BaseNode):
    """
    Translates spoken words in an audio file to English text.
    audio, stt, s2t, speech-to-text, translation, english
    Outputs the english translation of an audio file as text.
    """

    audio: AudioRef = Field(
        default=AudioRef(), description="The audio file to translate."
    )
    temperature: float = Field(
        default=0.0, description="The temperature to use for the translation."
    )

    async def process(self, context: ProcessingContext) -> str:
        audio_bytes = await context.asset_to_io(self.audio)
        response = await context.run_prediction(
            node_id=self._id,
            provider="openai",
            model="whisper-1",
            params={
                "file": base64.b64encode(audio_bytes.read()).decode(),
                "temperature": self.temperature,
                "translate": True,
            },
        )
        res = Translation(**response.json())

        return res.text



from typing import Any
from nodetool.workflows.property import Property
from nodetool.metadata.types import OutputSlot


from pydantic import BaseModel


class NodeMetadata(BaseModel):
    """
    Metadata for a node.
    """

    title: str
    description: str
    namespace: str
    node_type: str
    primary_field: str | None
    secondary_field: str | None
    layout: str
    properties: list[Property]
    outputs: list[OutputSlot]
    model_info: dict[str, Any]



from pydantic import Field
from nodetool.nodes.comfy.controlnet import PreprocessImage


class AnimeFace_SemSegPreprocessor(PreprocessImage):
    remove_background_using_abgr: bool = Field(
        default=True,
        description="Whether to remove the background.",
    )
    resolution: int = Field(
        default=512, description="The width of the image to generate.", ge=512, le=512
    )


class OneFormerCOCOSemSegPreprocessor(PreprocessImage):
    pass


class OneFormer_ADE20K_SemSegPreprocessor(PreprocessImage):
    pass


class UniformerSemSegPreprocessor(PreprocessImage):
    pass



from pydantic import BaseModel, Field
import nodetool.metadata.types
from nodetool.metadata.types import *
from nodetool.dsl.graph import GraphNode


class Add(GraphNode):
    a: int | float | nodetool.metadata.types.Tensor | GraphNode | tuple[GraphNode, str] = Field(default=0.0, description=None)
    b: int | float | nodetool.metadata.types.Tensor | GraphNode | tuple[GraphNode, str] = Field(default=0.0, description=None)
    @classmethod
    def get_node_type(cls): return "nodetool.math.Add"



class BinaryOperation(GraphNode):
    a: int | float | nodetool.metadata.types.Tensor | GraphNode | tuple[GraphNode, str] = Field(default=0.0, description=None)
    b: int | float | nodetool.metadata.types.Tensor | GraphNode | tuple[GraphNode, str] = Field(default=0.0, description=None)
    @classmethod
    def get_node_type(cls): return "nodetool.math.BinaryOperation"



class Cosine(GraphNode):
    angle_rad: float | int | nodetool.metadata.types.Tensor | GraphNode | tuple[GraphNode, str] = Field(default=0.0, description=None)
    @classmethod
    def get_node_type(cls): return "nodetool.math.Cosine"



class Divide(GraphNode):
    a: int | float | nodetool.metadata.types.Tensor | GraphNode | tuple[GraphNode, str] = Field(default=0.0, description=None)
    b: int | float | nodetool.metadata.types.Tensor | GraphNode | tuple[GraphNode, str] = Field(default=0.0, description=None)
    @classmethod
    def get_node_type(cls): return "nodetool.math.Divide"



class Modulus(GraphNode):
    a: int | float | nodetool.metadata.types.Tensor | GraphNode | tuple[GraphNode, str] = Field(default=0.0, description=None)
    b: int | float | nodetool.metadata.types.Tensor | GraphNode | tuple[GraphNode, str] = Field(default=0.0, description=None)
    @classmethod
    def get_node_type(cls): return "nodetool.math.Modulus"



class Multiply(GraphNode):
    a: int | float | nodetool.metadata.types.Tensor | GraphNode | tuple[GraphNode, str] = Field(default=0.0, description=None)
    b: int | float | nodetool.metadata.types.Tensor | GraphNode | tuple[GraphNode, str] = Field(default=0.0, description=None)
    @classmethod
    def get_node_type(cls): return "nodetool.math.Multiply"



class Power(GraphNode):
    base: float | int | nodetool.metadata.types.Tensor | GraphNode | tuple[GraphNode, str] = Field(default=1.0, description=None)
    exponent: float | int | nodetool.metadata.types.Tensor | GraphNode | tuple[GraphNode, str] = Field(default=2.0, description=None)
    @classmethod
    def get_node_type(cls): return "nodetool.math.Power"



class Sine(GraphNode):
    angle_rad: float | int | nodetool.metadata.types.Tensor | GraphNode | tuple[GraphNode, str] = Field(default=0.0, description=None)
    @classmethod
    def get_node_type(cls): return "nodetool.math.Sine"



class SqrtTensor(GraphNode):
    x: int | float | nodetool.metadata.types.Tensor | GraphNode | tuple[GraphNode, str] = Field(default=1.0, description=None)
    @classmethod
    def get_node_type(cls): return "nodetool.math.SqrtTensor"



class Subtract(GraphNode):
    a: int | float | nodetool.metadata.types.Tensor | GraphNode | tuple[GraphNode, str] = Field(default=0.0, description=None)
    b: int | float | nodetool.metadata.types.Tensor | GraphNode | tuple[GraphNode, str] = Field(default=0.0, description=None)
    @classmethod
    def get_node_type(cls): return "nodetool.math.Subtract"





CONTENT_TYPE_TO_EXTENSION = {
    "image/jpeg": "jpg",
    "image/png": "png",
    "image/gif": "gif",
    "image/svg+xml": "svg",
    "image/webp": "webp",
    "image/tiff": "tiff",
    "image/bmp": "bmp",
    "text/plain": "txt",
    "text/csv": "csv",
    "text/html": "html",
    "text/css": "css",
    "text/javascript": "js",
    "text/xml": "xml",
    "application/json": "json",
    "application/pdf": "pdf",
    "application/zip": "zip",
    "application/x-7z-compressed": "7z",
    "application/x-rar-compressed": "rar",
    "application/x-tar": "tar",
    "application/x-bzip": "bz",
    "application/x-bzip2": "bz2",
    "application/x-xz": "xz",
    "application/x-lzip": "lz",
    "application/x-lzma": "lzma",
    "application/x-compress": "Z",
    "application/x-gzip": "gz",
    "audio/midi": "midi",
    "audio/mpeg": "mp3",
    "audio/webm": "webm",
    "audio/ogg": "ogg",
    "audio/wav": "wav",
    "audio/aac": "aac",
    "audio/wave": "wave",
    "audio/x-wav": "wav",
    "audio/x-aiff": "aiff",
    "audio/x-flac": "flac",
    "audio/x-m4a": "m4a",
    "video/mpeg": "mpeg",
    "video/mp4": "mp4",
    "video/webm": "webm",
    "video/ogg": "ogg",
    "video/quicktime": "mov",
    "video/x-msvideo": "avi",
    "video/x-flv": "flv",
    "video/x-m4v": "m4v",
    "video/3gpp": "3gp",
    "video/3gpp2": "3g2",
}

EXTENSION_TO_CONTENT_TYPE = {
    value: key
    for key, value in CONTENT_TYPE_TO_EXTENSION.items()
}



from datetime import datetime
import os
import shutil
import aiofiles
from typing import IO, AsyncIterator, Iterator

from nodetool.storage.abstract_storage import AbstractStorage


class FileStorage(AbstractStorage):
    base_path: str
    base_url: str

    def __init__(self, base_path: str, base_url: str):
        self.base_path = base_path
        self.base_url = base_url
        os.makedirs(base_path, exist_ok=True)

    def get_base_url(self):
        return self.base_url

    def get_url(self, key: str):
        return f"{self.base_url}/{key}"

    def generate_presigned_url(
        self, client_method: str, object_name: str, expiration=3600 * 24 * 7
    ):
        return f"{self.base_url}/{object_name}"

    async def file_exists(self, file_name: str) -> bool:
        return os.path.isfile(os.path.join(self.base_path, file_name))

    async def get_mtime(self, key: str):
        try:
            mtime = os.path.getmtime(os.path.join(self.base_path, key))
            return datetime.fromtimestamp(mtime, tz=datetime.now().astimezone().tzinfo)
        except FileNotFoundError:
            return None

    async def download_stream(self, key: str) -> AsyncIterator[bytes]:
        with open(os.path.join(self.base_path, key), "rb") as f:
            while chunk := f.read(8192):
                yield chunk

    async def download(self, key: str, stream: IO):
        async with aiofiles.open(os.path.join(self.base_path, key), "rb") as f:
            async for chunk in f:
                stream.write(chunk)

    async def upload(self, key: str, content: IO) -> str:
        async with aiofiles.open(os.path.join(self.base_path, key), "wb") as f:
            while True:
                chunk = content.read(1024 * 1024)  # Read in 1MB chunks
                if not chunk:
                    break
                await f.write(chunk)

        return self.generate_presigned_url("get_object", key)

    async def delete(self, file_name: str):
        os.remove(os.path.join(self.base_path, file_name))



"""
Copyright (c) Microsoft Corporation.
Licensed under the MIT license.

Definition of TSV class
"""


import logging
import os
import os.path as op


def generate_lineidx(filein, idxout):
    idxout_tmp = idxout + '.tmp'
    with open(filein, 'r') as tsvin, open(idxout_tmp,'w') as tsvout:
        fsize = os.fstat(tsvin.fileno()).st_size
        fpos = 0
        while fpos!=fsize:
            tsvout.write(str(fpos)+"\n")
            tsvin.readline()
            fpos = tsvin.tell()
    os.rename(idxout_tmp, idxout)


def read_to_character(fp, c):
    result = []
    while True:
        s = fp.read(32)
        assert s != ''
        if c in s:
            result.append(s[: s.index(c)])
            break
        else:
            result.append(s)
    return ''.join(result)


class TSVFile(object):
    def __init__(self, tsv_file, generate_lineidx=False):
        self.tsv_file = tsv_file
        self.lineidx = op.splitext(tsv_file)[0] + '.lineidx'
        self._fp = None
        self._lineidx = None
        # the process always keeps the process which opens the file. 
        # If the pid is not equal to the currrent pid, we will re-open the file.
        self.pid = None
        # generate lineidx if not exist
        if not op.isfile(self.lineidx) and generate_lineidx:
            generate_lineidx(self.tsv_file, self.lineidx)

    def __del__(self):
        if self._fp:
            self._fp.close()

    def __str__(self):
        return "TSVFile(tsv_file='{}')".format(self.tsv_file)

    def __repr__(self):
        return str(self)

    def num_rows(self):
        self._ensure_lineidx_loaded()
        return len(self._lineidx)

    def seek(self, idx):
        self._ensure_tsv_opened()
        self._ensure_lineidx_loaded()
        try:
            pos = self._lineidx[idx]
        except:
            logging.info('{}-{}'.format(self.tsv_file, idx))
            raise
        self._fp.seek(pos)
        return [s.strip() for s in self._fp.readline().split('\t')]

    def seek_first_column(self, idx):
        self._ensure_tsv_opened()
        self._ensure_lineidx_loaded()
        pos = self._lineidx[idx]
        self._fp.seek(pos)
        return read_to_character(self._fp, '\t')

    def get_key(self, idx):
        return self.seek_first_column(idx)

    def __getitem__(self, index):
        return self.seek(index)

    def __len__(self):
        return self.num_rows()

    def _ensure_lineidx_loaded(self):
        if self._lineidx is None:
            logging.info('loading lineidx: {}'.format(self.lineidx))
            with open(self.lineidx, 'r') as fp:
                self._lineidx = [int(i.strip()) for i in fp.readlines()]

    def _ensure_tsv_opened(self):
        if self._fp is None:
            self._fp = open(self.tsv_file, 'r')
            self.pid = os.getpid()

        if self.pid != os.getpid():
            logging.info('re-open {} because the process id changed'.format(self.tsv_file))
            self._fp = open(self.tsv_file, 'r')
            self.pid = os.getpid()


class CompositeTSVFile():
    def __init__(self, file_list, seq_file, root='.'):
        if isinstance(file_list, str):
            self.file_list = load_list_file(file_list)
        else:
            assert isinstance(file_list, list)
            self.file_list = file_list

        self.seq_file = seq_file
        self.root = root
        self.initialized = False
        self.initialize()

    def get_key(self, index):
        idx_source, idx_row = self.seq[index]
        k = self.tsvs[idx_source].get_key(idx_row)
        return '_'.join([self.file_list[idx_source], k])

    def num_rows(self):
        return len(self.seq)

    def __getitem__(self, index):
        idx_source, idx_row = self.seq[index]
        return self.tsvs[idx_source].seek(idx_row)

    def __len__(self):
        return len(self.seq)

    def initialize(self):
        '''
        this function has to be called in init function if cache_policy is
        enabled. Thus, let's always call it in init funciton to make it simple.
        '''
        if self.initialized:
            return
        self.seq = []
        with open(self.seq_file, 'r') as fp:
            for line in fp:
                parts = line.strip().split('\t')
                self.seq.append([int(parts[0]), int(parts[1])])
        self.tsvs = [TSVFile(op.join(self.root, f)) for f in self.file_list]
        self.initialized = True


def load_list_file(fname):
    with open(fname, 'r') as fp:
        lines = fp.readlines()
    result = [line.strip() for line in lines]
    if len(result) > 0 and result[-1] == '':
        result = result[:-1]
    return result





# -*- coding: utf-8 -*-
# Copyright (c) Facebook, Inc. and its affiliates.

import logging
from typing import Dict, List
import torch
from torch import nn

from custom_detectron2.config import configurable
from custom_detectron2.structures import ImageList

from ..postprocessing import detector_postprocess, sem_seg_postprocess
from .build import META_ARCH_REGISTRY
from .rcnn import GeneralizedRCNN
from .semantic_seg import build_sem_seg_head

__all__ = ["PanopticFPN"]


@META_ARCH_REGISTRY.register()
class PanopticFPN(GeneralizedRCNN):
    """
    Implement the paper :paper:`PanopticFPN`.
    """

    @configurable
    def __init__(
        self,
        *,
        sem_seg_head: nn.Module,
        combine_overlap_thresh: float = 0.5,
        combine_stuff_area_thresh: float = 4096,
        combine_instances_score_thresh: float = 0.5,
        **kwargs,
    ):
        """
        NOTE: this interface is experimental.

        Args:
            sem_seg_head: a module for the semantic segmentation head.
            combine_overlap_thresh: combine masks into one instances if
                they have enough overlap
            combine_stuff_area_thresh: ignore stuff areas smaller than this threshold
            combine_instances_score_thresh: ignore instances whose score is
                smaller than this threshold

        Other arguments are the same as :class:`GeneralizedRCNN`.
        """
        super().__init__(**kwargs)
        self.sem_seg_head = sem_seg_head
        # options when combining instance & semantic outputs
        self.combine_overlap_thresh = combine_overlap_thresh
        self.combine_stuff_area_thresh = combine_stuff_area_thresh
        self.combine_instances_score_thresh = combine_instances_score_thresh

    @classmethod
    def from_config(cls, cfg):
        ret = super().from_config(cfg)
        ret.update(
            {
                "combine_overlap_thresh": cfg.MODEL.PANOPTIC_FPN.COMBINE.OVERLAP_THRESH,
                "combine_stuff_area_thresh": cfg.MODEL.PANOPTIC_FPN.COMBINE.STUFF_AREA_LIMIT,
                "combine_instances_score_thresh": cfg.MODEL.PANOPTIC_FPN.COMBINE.INSTANCES_CONFIDENCE_THRESH,  # noqa
            }
        )
        ret["sem_seg_head"] = build_sem_seg_head(cfg, ret["backbone"].output_shape())
        logger = logging.getLogger(__name__)
        if not cfg.MODEL.PANOPTIC_FPN.COMBINE.ENABLED:
            logger.warning(
                "PANOPTIC_FPN.COMBINED.ENABLED is no longer used. "
                " model.inference(do_postprocess=) should be used to toggle postprocessing."
            )
        if cfg.MODEL.PANOPTIC_FPN.INSTANCE_LOSS_WEIGHT != 1.0:
            w = cfg.MODEL.PANOPTIC_FPN.INSTANCE_LOSS_WEIGHT
            logger.warning(
                "PANOPTIC_FPN.INSTANCE_LOSS_WEIGHT should be replaced by weights on each ROI head."
            )

            def update_weight(x):
                if isinstance(x, dict):
                    return {k: v * w for k, v in x.items()}
                else:
                    return x * w

            roi_heads = ret["roi_heads"]
            roi_heads.box_predictor.loss_weight = update_weight(roi_heads.box_predictor.loss_weight)
            roi_heads.mask_head.loss_weight = update_weight(roi_heads.mask_head.loss_weight)
        return ret

    def forward(self, batched_inputs):
        """
        Args:
            batched_inputs: a list, batched outputs of :class:`DatasetMapper`.
                Each item in the list contains the inputs for one image.

                For now, each item in the list is a dict that contains:

                * "image": Tensor, image in (C, H, W) format.
                * "instances": Instances
                * "sem_seg": semantic segmentation ground truth.
                * Other information that's included in the original dicts, such as:
                  "height", "width" (int): the output resolution of the model, used in inference.
                  See :meth:`postprocess` for details.

        Returns:
            list[dict]:
                each dict has the results for one image. The dict contains the following keys:

                * "instances": see :meth:`GeneralizedRCNN.forward` for its format.
                * "sem_seg": see :meth:`SemanticSegmentor.forward` for its format.
                * "panoptic_seg": See the return value of
                  :func:`combine_semantic_and_instance_outputs` for its format.
        """
        if not self.training:
            return self.inference(batched_inputs)
        images = self.preprocess_image(batched_inputs)
        features = self.backbone(images.tensor)

        assert "sem_seg" in batched_inputs[0]
        gt_sem_seg = [x["sem_seg"].to(self.device) for x in batched_inputs]
        gt_sem_seg = ImageList.from_tensors(
            gt_sem_seg,
            self.backbone.size_divisibility,
            self.sem_seg_head.ignore_value,
            self.backbone.padding_constraints,
        ).tensor
        sem_seg_results, sem_seg_losses = self.sem_seg_head(features, gt_sem_seg)

        gt_instances = [x["instances"].to(self.device) for x in batched_inputs]
        proposals, proposal_losses = self.proposal_generator(images, features, gt_instances)
        detector_results, detector_losses = self.roi_heads(
            images, features, proposals, gt_instances
        )

        losses = sem_seg_losses
        losses.update(proposal_losses)
        losses.update(detector_losses)
        return losses

    def inference(self, batched_inputs: List[Dict[str, torch.Tensor]], do_postprocess: bool = True):
        """
        Run inference on the given inputs.

        Args:
            batched_inputs (list[dict]): same as in :meth:`forward`
            do_postprocess (bool): whether to apply post-processing on the outputs.

        Returns:
            When do_postprocess=True, see docs in :meth:`forward`.
            Otherwise, returns a (list[Instances], list[Tensor]) that contains
            the raw detector outputs, and raw semantic segmentation outputs.
        """
        images = self.preprocess_image(batched_inputs)
        features = self.backbone(images.tensor)
        sem_seg_results, sem_seg_losses = self.sem_seg_head(features, None)
        proposals, _ = self.proposal_generator(images, features, None)
        detector_results, _ = self.roi_heads(images, features, proposals, None)

        if do_postprocess:
            processed_results = []
            for sem_seg_result, detector_result, input_per_image, image_size in zip(
                sem_seg_results, detector_results, batched_inputs, images.image_sizes
            ):
                height = input_per_image.get("height", image_size[0])
                width = input_per_image.get("width", image_size[1])
                sem_seg_r = sem_seg_postprocess(sem_seg_result, image_size, height, width)
                detector_r = detector_postprocess(detector_result, height, width)

                processed_results.append({"sem_seg": sem_seg_r, "instances": detector_r})

                panoptic_r = combine_semantic_and_instance_outputs(
                    detector_r,
                    sem_seg_r.argmax(dim=0),
                    self.combine_overlap_thresh,
                    self.combine_stuff_area_thresh,
                    self.combine_instances_score_thresh,
                )
                processed_results[-1]["panoptic_seg"] = panoptic_r
            return processed_results
        else:
            return detector_results, sem_seg_results


def combine_semantic_and_instance_outputs(
    instance_results,
    semantic_results,
    overlap_threshold,
    stuff_area_thresh,
    instances_score_thresh,
):
    """
    Implement a simple combining logic following
    "combine_semantic_and_instance_predictions.py" in panopticapi
    to produce panoptic segmentation outputs.

    Args:
        instance_results: output of :func:`detector_postprocess`.
        semantic_results: an (H, W) tensor, each element is the contiguous semantic
            category id

    Returns:
        panoptic_seg (Tensor): of shape (height, width) where the values are ids for each segment.
        segments_info (list[dict]): Describe each segment in `panoptic_seg`.
            Each dict contains keys "id", "category_id", "isthing".
    """
    panoptic_seg = torch.zeros_like(semantic_results, dtype=torch.int32)

    # sort instance outputs by scores
    sorted_inds = torch.argsort(-instance_results.scores)

    current_segment_id = 0
    segments_info = []

    instance_masks = instance_results.pred_masks.to(dtype=torch.bool, device=panoptic_seg.device)

    # Add instances one-by-one, check for overlaps with existing ones
    for inst_id in sorted_inds:
        score = instance_results.scores[inst_id].item()
        if score < instances_score_thresh:
            break
        mask = instance_masks[inst_id]  # H,W
        mask_area = mask.sum().item()

        if mask_area == 0:
            continue

        intersect = (mask > 0) & (panoptic_seg > 0)
        intersect_area = intersect.sum().item()

        if intersect_area * 1.0 / mask_area > overlap_threshold:
            continue

        if intersect_area > 0:
            mask = mask & (panoptic_seg == 0)

        current_segment_id += 1
        panoptic_seg[mask] = current_segment_id
        segments_info.append(
            {
                "id": current_segment_id,
                "isthing": True,
                "score": score,
                "category_id": instance_results.pred_classes[inst_id].item(),
                "instance_id": inst_id.item(),
            }
        )

    # Add semantic results to remaining empty areas
    semantic_labels = torch.unique(semantic_results).cpu().tolist()
    for semantic_label in semantic_labels:
        if semantic_label == 0:  # 0 is a special "thing" class
            continue
        mask = (semantic_results == semantic_label) & (panoptic_seg == 0)
        mask_area = mask.sum().item()
        if mask_area < stuff_area_thresh:
            continue

        current_segment_id += 1
        panoptic_seg[mask] = current_segment_id
        segments_info.append(
            {
                "id": current_segment_id,
                "isthing": False,
                "category_id": semantic_label,
                "area": mask_area,
            }
        )

    return panoptic_seg, segments_info



import os
from fastapi.exceptions import RequestValidationError
from fastapi.responses import JSONResponse
from fastapi.staticfiles import StaticFiles
from nodetool.api import prediction
from nodetool.common.environment import Environment

from fastapi import APIRouter, FastAPI, Request
from fastapi.middleware.cors import CORSMiddleware
from uvicorn import run as uvicorn

from . import (
    asset,
    job,
    auth,
    message,
    node,
    storage,
    task,
    workflow,
    model,
)

DEFAULT_ROUTERS = [
    asset.router,
    job.router,
    auth.router,
    message.router,
    node.router,
    prediction.router,
    workflow.router,
    storage.router,
    task.router,
    model.router,
]


def create_app(
    origins: list[str] = ["*"],
    routers: list[APIRouter] = DEFAULT_ROUTERS,
    static_folder: str | None = None,
):
    app = FastAPI()
    app.add_middleware(
        CORSMiddleware,
        allow_origins=origins,
        allow_credentials=True,
        allow_methods=["*"],
        allow_headers=["*"],
        expose_headers=["*"],
        max_age=3600,
    )

    for router in routers:
        app.include_router(router)

    if not Environment.is_production():

        @app.exception_handler(RequestValidationError)
        async def validation_exception_handler(
            request: Request, exc: RequestValidationError
        ):
            print(f"Request validation error: {exc}")
            return JSONResponse({"detail": exc.errors()}, status_code=422)

    @app.get("/health")
    async def health_check() -> str:
        return "OK"

    if static_folder:
        app.mount("/", StaticFiles(directory=static_folder, html=True), name="static")

    return app


def run_uvicorn_server(app: str, host: str, port: int, reload: bool) -> None:
    """
    Starts api using Uvicorn with the specified configuration.

    Args:
        app: The app to run.
        host: The host to run on.
        port: The port to run on.
        reload: Whether to reload the server on changes.
    """
    current_dir = os.path.dirname(os.path.realpath(__file__))
    parent_dir = os.path.dirname(current_dir)
    uvicorn(
        app=app,
        host=host,
        port=port,
        reload=reload,
        reload_dirs=[parent_dir] if reload else None,
        reload_includes=["*.py"] if reload else None,
    )
